# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'exception_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QFrame,
    QHeaderView,
    QLabel,
    QSizePolicy,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)


class Ui_ExceptionDialog(object):
    def setupUi(self, ExceptionDialog):
        if not ExceptionDialog.objectName():
            ExceptionDialog.setObjectName("ExceptionDialog")
        ExceptionDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(ExceptionDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.exception_label = QLabel(ExceptionDialog)
        self.exception_label.setObjectName("exception_label")

        self.verticalLayout.addWidget(self.exception_label)

        self.line = QFrame(ExceptionDialog)
        self.line.setObjectName("line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.verticalLayout.addWidget(self.line)

        self.exception_tree = QTreeWidget(ExceptionDialog)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setText(0, "1")
        self.exception_tree.setHeaderItem(__qtreewidgetitem)
        self.exception_tree.setObjectName("exception_tree")
        self.exception_tree.setHeaderHidden(True)

        self.verticalLayout.addWidget(self.exception_tree)

        self.buttonBox = QDialogButtonBox(ExceptionDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(ExceptionDialog)
        self.buttonBox.accepted.connect(ExceptionDialog.accept)
        self.buttonBox.rejected.connect(ExceptionDialog.reject)

        QMetaObject.connectSlotsByName(ExceptionDialog)

    # setupUi

    def retranslateUi(self, ExceptionDialog):
        ExceptionDialog.setWindowTitle(
            QCoreApplication.translate("ExceptionDialog", "Dialog", None)
        )
        self.exception_label.setText(
            QCoreApplication.translate("ExceptionDialog", "An error occured", None)
        )

    # retranslateUi



import decimal
from typing import SupportsFloat, TYPE_CHECKING

if TYPE_CHECKING:
    from caqtus.device.sequencer import TimeStep

_ns = decimal.Decimal("1e-9")


def duration_to_ticks(duration: SupportsFloat, time_step: "TimeStep") -> int:
    """Returns the nearest number of ticks for a given duration and time step.

    Args:
        duration: The duration in seconds.
        time_step: The time step in nanoseconds.
    """

    dt = time_step * _ns

    rounded = round(float(duration) / float(dt))
    if not isinstance(rounded, int):
        raise TypeError(f"Expected integer number of ticks, got {rounded}")
    return rounded



from __future__ import annotations

from collections.abc import Sequence
from typing import Optional, Literal, TypeAlias, Any

import numpy as np
import polars
import pyqtgraph
from PyQt6 import QtGui, QtCore
from PyQt6.QtWidgets import QWidget
from caqtus.analysis.stats import compute_stats_average, get_nominal_value, get_error
from caqtus.analysis.units import extract_unit
from caqtus.types.units import dimensionless, Unit
from pyqtgraph import PlotWidget

from .errorbar_visualizer_ui import Ui_ErrorBarVisualizerCreator
from ..visualizer_creator import ViewCreator, DataView

pyqtgraph.setConfigOptions(antialias=True)


class ErrorBarViewCreator(QWidget, ViewCreator, Ui_ErrorBarVisualizerCreator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._setup_ui()

    def _setup_ui(self) -> None:
        self.setupUi(self)

    def create_view(self) -> ErrorBarView:
        x = self._x_axis_line_edit.text()
        y = self._y_axis_line_edit.text()
        hue = text if (text := self._hue_line_edit.text()) else None
        if hue is None:
            view = ErrorBarView(x, y)
        else:
            view = HueErrorBarView(x, y, hue)
        font = QtGui.QFont()
        font.setPixelSize(25)
        view.getAxis("bottom").setStyle(tickFont=font)
        view.getAxis("left").setStyle(tickFont=font)
        view.getAxis("bottom").label.setFont(font)
        view.getAxis("left").label.setFont(font)

        pen = pyqtgraph.mkPen(color=(255, 255, 255), width=2)
        view.plotItem.getAxis("bottom").setPen(pen)
        view.plotItem.getAxis("left").setPen(pen)
        view.getAxis("bottom").setTextPen(pen)
        view.getAxis("left").setTextPen(pen)
        return view


WhichAxis: TypeAlias = Literal["left", "right", "top", "bottom"]


class HueErrorBarView(PlotWidget, DataView):
    def __init__(self, x: str, y: str, hue: str, *args, **kwargs):
        PlotWidget.__init__(self, *args, background=(0, 0, 0, 0), **kwargs)

        self._x_var = x
        self._y_var = y
        self._hue = hue

        self._stats: Optional[polars.DataFrame] = None

        self._error_bar_plotters: dict[Any, FilledErrorBarPlotter] = {}
        self._compute_stats_thread: Optional[ComputeStatsThread] = None

        self.update_plot()

    def update_plot(self) -> None:
        stats = self._stats
        if stats is None or len(stats) == 0:
            x_unit = None
            y_unit = None
            hue_unit = None
        else:
            for hue, group in stats.groupby(self._hue):
                if hue not in self._error_bar_plotters:
                    self._error_bar_plotters[hue] = FilledErrorBarPlotter()
                    for item in self._error_bar_plotters[hue].get_items():
                        self.addItem(item)
                x_series, x_unit = extract_unit(group[self._x_var])
                y_series, y_unit = extract_unit(group[f"{self._y_var}.mean"])
                error_series, _ = extract_unit(group[f"{self._y_var}.sem"])
                x_values = np.array(x_series.to_numpy())
                y_values = np.array(y_series.to_numpy())
                error_values = np.array(error_series.to_numpy())
                self._error_bar_plotters[hue].set_data(
                    x=x_values,
                    y=y_values,
                    error=error_values,
                )

        self.update_axis_labels(x_unit, y_unit)

    def update_axis_labels(
        self, x_unit: Optional[Unit], y_unit: Optional[Unit]
    ) -> None:
        self.setLabel("bottom", format_label(self._x_var, x_unit))
        self.setLabel("left", format_label(self._y_var, y_unit))

    def update_data(self, dataframe: Optional[polars.DataFrame]) -> None:
        if self._compute_stats_thread is not None:
            if self._compute_stats_thread.isRunning():
                return
        self._compute_stats_thread = ComputeStatsThread(
            self, dataframe, self._x_var, self._y_var, hue=self._hue
        )
        self._compute_stats_thread.finished.connect(self.update_plot)  # type: ignore
        self._compute_stats_thread.start()


class ErrorBarView(PlotWidget, DataView):
    def __init__(self, x: str, y: str, *args, **kwargs):
        PlotWidget.__init__(self, *args, background=(0, 0, 0, 0), **kwargs)

        self._x_var = x
        self._y_var = y

        self._stats: Optional[polars.DataFrame] = None

        self._error_bar_plotter = FilledErrorBarPlotter()
        self._compute_stats_thread: Optional[ComputeStatsThread] = None

        self._setup_ui()

    def _setup_ui(self):
        for item in self._error_bar_plotter.get_items():
            self.addItem(item)
        self.update_plot()

    def update_plot(self) -> None:
        stats = self._stats
        if stats is None:
            x_unit = None
            y_unit = None
            x_values = np.array([])
            y_values = np.array([])
            error_values = np.array([])
        else:
            x_series, x_unit = extract_unit(stats[self._x_var])
            y_series, y_unit = extract_unit(stats[self._y_var])
            nominal_series = get_nominal_value(y_series)
            error_series = get_error(y_series)
            x_values = x_series.to_numpy()
            y_values = nominal_series.to_numpy()
            error_values = error_series.to_numpy()

        self.update_axis_labels(x_unit, y_unit)
        self._error_bar_plotter.set_data(
            x=x_values,
            y=y_values,
            error=error_values,
        )

    def update_axis_labels(
        self, x_unit: Optional[Unit], y_unit: Optional[Unit]
    ) -> None:
        self.setLabel("bottom", format_label(self._x_var, x_unit))
        self.setLabel("left", format_label(self._y_var, y_unit))

    def update_data(self, dataframe: Optional[polars.DataFrame]) -> None:
        if self._compute_stats_thread is not None:
            if self._compute_stats_thread.isRunning():
                return
        self._compute_stats_thread = ComputeStatsThread(
            self,
            dataframe,
            self._x_var,
            self._y_var,
        )
        self._compute_stats_thread.finished.connect(self.update_plot)  # type: ignore
        self._compute_stats_thread.start()


class ComputeStatsThread(QtCore.QThread):
    def __init__(
        self,
        parent: ErrorBarView,
        dataframe: Optional[polars.DataFrame],
        x_var: str,
        y_var: str,
        hue: Optional[str] = None,
    ):
        super().__init__(parent=parent)
        self._parent = parent
        self._dataframe = dataframe
        self._x_var = x_var
        self._y_var = y_var
        self._hue = hue

    def run(self) -> None:
        if self._dataframe is not None and len(self._dataframe) > 0:
            group_by = [self._x_var] if self._hue is None else [self._hue, self._x_var]
            self._parent._stats = compute_stats_average(
                self._dataframe, self._y_var, group_by
            )
        else:
            self._parent.stats = None


def format_label(label: str, unit: Optional[Unit]) -> str:
    if unit is None:
        return label
    else:
        if unit == dimensionless:
            return label
        else:
            return f"{label} [{unit:~}]"


class ErrorBarPlotter:
    def __init__(self):
        self._error_bar_item = pyqtgraph.ErrorBarItem(
            x=np.array([]), y=np.array([]), height=np.array([])
        )

    def set_data(self, x, y, error) -> None:
        self._error_bar_item.setData(x=x, y=y, height=error)

    def get_items(self) -> Sequence[pyqtgraph.GraphicsObject]:
        return [self._error_bar_item]


class FilledErrorBarPlotter:
    def __init__(self):
        self._top_curve = pyqtgraph.PlotCurveItem(x=np.array([]), y=np.array([]))
        self._bottom_curve = pyqtgraph.PlotCurveItem(x=np.array([]), y=np.array([]))
        self._middle_curve = pyqtgraph.PlotCurveItem(
            x=np.array([], dtype=float),
            y=np.array([], dtype=float),
            pen=pyqtgraph.mkPen("w", width=2),
        )
        self._fill = pyqtgraph.FillBetweenItem(
            self._top_curve, self._bottom_curve, brush=0.2
        )

    def set_data(self, x, y, error) -> None:
        # Here 1.96 refers to the 95% confidence interval
        self._top_curve.setData(x, y + error * 1.96)
        self._bottom_curve.setData(x, y - error * 1.96)
        self._middle_curve.setData(x, y)

    def get_items(self) -> Sequence[pyqtgraph.GraphicsObject]:
        return [self._fill, self._middle_curve]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'scatter_view.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QHBoxLayout, QLabel, QLineEdit,
    QSizePolicy, QSpacerItem, QToolButton, QVBoxLayout,
    QWidget)

class Ui_ScatterView(object):
    def setupUi(self, ScatterView):
        if not ScatterView.objectName():
            ScatterView.setObjectName(u"ScatterView")
        ScatterView.resize(400, 300)
        self.verticalLayout = QVBoxLayout(ScatterView)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label = QLabel(ScatterView)
        self.label.setObjectName(u"label")

        self.horizontalLayout.addWidget(self.label)

        self.x_line_edit = QLineEdit(ScatterView)
        self.x_line_edit.setObjectName(u"x_line_edit")

        self.horizontalLayout.addWidget(self.x_line_edit)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.label_2 = QLabel(ScatterView)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout.addWidget(self.label_2)

        self.y_line_edit = QLineEdit(ScatterView)
        self.y_line_edit.setObjectName(u"y_line_edit")

        self.horizontalLayout.addWidget(self.y_line_edit)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.apply_button = QToolButton(ScatterView)
        self.apply_button.setObjectName(u"apply_button")

        self.horizontalLayout.addWidget(self.apply_button)

        self.settings_button = QToolButton(ScatterView)
        self.settings_button.setObjectName(u"settings_button")

        self.horizontalLayout.addWidget(self.settings_button)


        self.verticalLayout.addLayout(self.horizontalLayout)


        self.retranslateUi(ScatterView)

        QMetaObject.connectSlotsByName(ScatterView)
    # setupUi

    def retranslateUi(self, ScatterView):
        ScatterView.setWindowTitle(QCoreApplication.translate("ScatterView", u"Form", None))
        self.label.setText(QCoreApplication.translate("ScatterView", u"x:", None))
        self.label_2.setText(QCoreApplication.translate("ScatterView", u"y:", None))
        self.apply_button.setText(QCoreApplication.translate("ScatterView", u"...", None))
        self.settings_button.setText(QCoreApplication.translate("ScatterView", u"...", None))
    # retranslateUi




from typing import TypeVar

import attrs

from caqtus.types.image.roi import RectangularROI
from ._runtime import Camera
from ..configuration import DeviceConfiguration


@attrs.define
class CameraConfiguration[C: Camera](DeviceConfiguration[C]):
    """Contains the necessary information about a camera.

    Attributes:
        roi: The region of interest to keep for the images taken by the camera for all
            acquisitions.
    """

    roi: RectangularROI = attrs.field(
        validator=attrs.validators.instance_of(RectangularROI),
        on_setattr=attrs.setters.validate,
    )


CameraConfigurationType = TypeVar("CameraConfigurationType", bound=CameraConfiguration)



from .manager import (
    ExperimentManager,
    Procedure,
    LocalExperimentManager,
    BoundProcedure,
    ProcedureNotActiveError,
)
from .remote_manager import (
    RemoteExperimentManagerServer,
    RemoteExperimentManagerClient,
    ExperimentManagerProxy,
    ProcedureProxy,
)
from ._configuration import (
    ExperimentManagerConnection,
    LocalExperimentManagerConfiguration,
    RemoteExperimentManagerConfiguration,
)

__all__ = [
    "ExperimentManager",
    "Procedure",
    "LocalExperimentManager",
    "BoundProcedure",
    "ProcedureNotActiveError",
    "RemoteExperimentManagerServer",
    "RemoteExperimentManagerClient",
    "ExperimentManagerProxy",
    "ProcedureProxy",
    "ExperimentManagerConnection",
    "LocalExperimentManagerConfiguration",
    "RemoteExperimentManagerConfiguration",
]



"""This module defines a custom dtype for polars dataframes that can be used to
represent quantities with units. This
dtype is called QuantityDType and is a struct with two fields: magnitude and units.
This module also defines functions
to add or remove units from series.
"""

from typing import Optional, Mapping
from typing_extensions import deprecated

import polars
from caqtus.types.units import Unit, Quantity

MAGNITUDE_FIELD = "magnitude"
UNITS_FIELD = "units"


@polars.api.register_expr_namespace("quantity")
class QuantityExpressions:
    def __init__(self, expr: polars.Expr):
        self._expr = expr

    def magnitude(self) -> polars.Expr:
        return self._expr.struct.field(MAGNITUDE_FIELD)

    def unit(self) -> polars.Expr:
        return self._expr.struct.field(UNITS_FIELD)


@polars.api.register_series_namespace("quantity")
class QuantitySeries:
    def __init__(self, s: polars.Series):
        self._s = s

    def extract_unit(self) -> tuple[polars.Series, Optional[str]]:
        """Break the series into a magnitude series and a unit.

        If the series has a quantity data type, this will attempt to convert all
        magnitudes to a given unit. It will then
        return a series of magnitudes only and their unit. If the series is any other
        dtype, it will be returned unchanged.
        """

        series = self._s
        if is_quantity_dtype(series.dtype):
            all_units = series.struct.field(UNITS_FIELD).unique().to_list()
            if len(all_units) == 1:
                unit = Unit(all_units[0])
            else:
                raise NotImplementedError(
                    f"Series {series.name} is expressed in several units: {all_units}"
                )
            magnitude = series.struct.field(MAGNITUDE_FIELD).alias(series.name)
        else:
            unit = None
            magnitude = series
        if unit is None:
            return magnitude, None
        else:
            return magnitude, format(unit, "~")


@polars.api.register_dataframe_namespace("quantity")
class QuantityDataframe:
    def __init__(self, df: polars.DataFrame):
        self._df = df

    def extract_units(self) -> tuple[polars.DataFrame, Mapping[str, Optional[str]]]:
        """Break a dataframe potentially containing quantities into its magnitude and
        unit components."""

        dataframe = self._df

        column_units = {}
        columns_magnitudes = {}
        for column in dataframe.columns:
            magnitude, unit = extract_unit(dataframe[column])
            columns_magnitudes[column] = magnitude
            column_units[column] = unit
        return polars.DataFrame(columns_magnitudes), column_units


def is_quantity_dtype(dtype: polars.DataType) -> bool:
    """Check if a dtype is a QuantityDType.

    Args:
        dtype: the dtype to check.

    Returns:
        True if the dtype is a polars.Struct with two fields, magnitude and units,
        False otherwise.
    """

    if isinstance(dtype, polars.Struct):
        if len(dtype.fields) == 2:
            if (
                dtype.fields[0].name == MAGNITUDE_FIELD
                and dtype.fields[1].name == UNITS_FIELD
            ):
                return True
    return False


def add_unit(series: polars.Series, unit: Optional[Unit]) -> polars.Series:
    """Add a unit to a series, if it is not None.

    Args:
        series: the series to which the unit should be added.
        unit: the unit to add. If None, the series is returned unchanged.

    Returns:
        A new series with the unit added. If the unit is None, the series is returned
        unchanged and has the same dtype.
        If the unit is not None, the series is returned with a quantity dtype having
        the unit as a categorical and
        the magnitude with the same dtype as the original series.
    """

    if unit is None:
        return series
    else:
        return polars.Series(
            series.name,
            [
                series.alias(MAGNITUDE_FIELD),
                polars.Series(
                    UNITS_FIELD, (str(unit),) * len(series), dtype=polars.Categorical
                ),
            ],
            dtype=polars.Struct,
        )


@deprecated("Use series.quantity.extract_unit() instead.")
def extract_unit(
    series: polars.Series,
) -> tuple[polars.Series, Optional[Unit]]:
    return series.quantity.extract_unit()


@deprecated("Use dataframe.quantity.extract_units() instead.")
def extract_units(
    dataframe: polars.DataFrame,
) -> tuple[polars.DataFrame, Mapping[str, Optional[str]]]:
    return dataframe.quantity.extract_units()


def convert_to_unit(
    series: polars.Series, target_unit: Optional[Unit]
) -> polars.Series:
    """Convert a series to a given unit.

    Args:
        series: the series to convert. If it has a quantity dtype, target_unit must
        be not be None. If it has any other
            dtype, target_unit must be None.
        target_unit: the unit to convert to.

    Returns:
        A new series with the same name as the original series. If unit is None,
        the series is returned unchanged. If
        the series has a quantity dtype, the series is returned with the same dtype,
        but all magnitudes converted to
        the target unit.
    """

    if target_unit is None:
        if is_quantity_dtype(series.dtype):
            raise ValueError(
                f"Series {series.name} is expressed in unit {extract_unit(series)[1]} "
                f"and target_unit is None"
            )
        else:
            return series

    magnitudes, unit = extract_unit(series)
    if unit is None:
        raise ValueError(
            f"Series {series.name} has no unit and needs to be converted to "
            f"{target_unit:~}"
        )
    else:
        quantity = Quantity(magnitudes.to_numpy(), unit).to(target_unit)
        return add_unit(polars.Series(series.name, quantity.magnitude), target_unit)


def magnitude_in_unit(series: polars.Series, unit: Optional[Unit]) -> polars.Series:
    """Return the magnitude of a series in a given unit.

    Args:
        series: the series to convert. It should have dtype QuantityDType.
        unit: the unit to convert to. If None and the series has a not a quantity
        dtype, the series is returned
            unchanged.

    Raises:
        ValueError: if the series has a quantity dtype and unit is None.

    Returns:
        A new series with the same name as the original series, with dtype Float64
        and all magnitudes
        converted to the target unit.
    """

    if unit is None:
        if is_quantity_dtype(series.dtype):
            raise ValueError(
                f"Series {series.name} is expressed in unit {extract_unit(series)[1]} "
                f"and target_unit is None"
            )
        else:
            return series
    return (
        convert_to_unit(series, unit).struct.field(MAGNITUDE_FIELD).alias(series.name)
    )


def with_columns_expressed_in_units(
    dataframe: polars.DataFrame, column_units: Mapping[str, Optional[Unit]]
) -> polars.DataFrame:
    """Compute the magnitude of columns in a dataframe to given units.

    Args:
        dataframe: the dataframe to convert.
        column_units: a mapping from column names to units.

    Returns:
        A new dataframe with the same columns as the original dataframe, but with all
        columns containing the magnitude
        of the original columns in the requested units.
    """

    return dataframe.with_columns(
        **{
            column: magnitude_in_unit(dataframe[column], column_units[column])
            for column in column_units
        }
    )


def with_units_added_to_columns(
    dataframe: polars.DataFrame, column_units: Mapping[str, Optional[Unit]]
) -> polars.DataFrame:
    """Add units to columns in a dataframe.

    Args:
        dataframe: the dataframe to convert.
        column_units: a mapping from column names to units. If the unit is None,
        the column is returned unchanged.

    Returns:
        A new dataframe with the same columns as the original dataframe, but with all
        columns containing the magnitude
        of the original columns with the requested units.
    """

    return dataframe.with_columns(
        **{
            column: add_unit(dataframe[column], column_units[column])
            for column in column_units
        }
    )



"""Contains functions for computing timing of a sequencer."""

import decimal
import math
from collections.abc import Iterable
from itertools import accumulate
from typing import SupportsFloat

ns = decimal.Decimal("1e-9")


def start_tick(start_time: SupportsFloat, time_step: SupportsFloat) -> int:
    """Returns the included first tick index of the step starting at start_time."""

    return math.ceil(float(start_time) / float(time_step))


def stop_tick(stop_time: SupportsFloat, time_step: SupportsFloat) -> int:
    """Returns the excluded last tick index of the step ending at stop_time."""

    return math.ceil(float(stop_time) / float(time_step))


def number_ticks(
    start_time: SupportsFloat, stop_time: SupportsFloat, time_step: SupportsFloat
) -> int:
    """Returns the number of ticks between start_time and stop_time.

    Args:
        start_time: The start time in seconds.
        stop_time: The stop time in seconds.
        time_step: The time step in seconds.
    """

    return stop_tick(stop_time, time_step) - start_tick(start_time, time_step)


def get_step_bounds(step_durations: Iterable[float]) -> list[float]:
    """Returns the time at which each step starts from their durations.

    For an iterable of step durations [d_0, d_1, ..., d_n], the step starts are
    [0, d_0, d_0 + d_1, ..., d_0 + ... + d_n]. It has one more element than the
    iterable of step durations, with the last element being the total duration.
    """

    return [0.0] + list((accumulate(step_durations)))



from __future__ import annotations

import datetime
from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional, Self

import attrs
import polars

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.types.data import DataLabel
from caqtus.types.iteration import IterationConfiguration, Unknown
from caqtus.types.parameter import ParameterNamespace
from caqtus.types.timelane import TimeLanes
from caqtus.types.variable_name import DottedVariableName
from .shot import Shot
from .._path import PureSequencePath
from .._return_or_raise import unwrap
from .._sequence_collection import PathIsNotSequenceError
from .._state import State

if TYPE_CHECKING:
    from .._experiment_session import ExperimentSession
    from caqtus.analysis.loading import DataImporter


def _convert_to_path(path: PureSequencePath | str) -> PureSequencePath:
    if isinstance(path, str):
        return PureSequencePath(path)
    return path


@attrs.frozen(eq=False, order=False)
class Sequence:
    """Represent a sequence in the experiment session.

    Sequence objects can be obtained by calling :meth:`ExperimentSession.get_sequence`.
    The returned sequence object is bound to the session and is only valid in the
    context where the session is active.

    Args:
        path: The path of the sequence.
        session: The session to which the sequence belongs.
            The sequence is bound to the session and is only valid in the context where
            the session is active.
    """

    path: PureSequencePath = attrs.field(converter=_convert_to_path)
    session: ExperimentSession

    def __attrs_post_init__(self):
        is_sequence = unwrap(self.session.sequences.is_sequence(self.path))
        if not is_sequence:
            raise PathIsNotSequenceError(self.path)

    @classmethod
    def create(
        cls,
        path: PureSequencePath,
        iteration_configuration: IterationConfiguration,
        time_lanes: TimeLanes,
        session: ExperimentSession,
    ) -> Self:
        """Create a new sequence in the session.

        Args:
            path: The path at which to create the sequence.
            iteration_configuration: How the sequence parameters should be iterated
                over.
            time_lanes: How the shots should be run.
            session: The session in which the sequence should be created.
                The session must be active.
        """

        session.sequences.create(path, iteration_configuration, time_lanes)
        return cls(path, session)

    def __str__(self) -> str:
        return str(self.path)

    def __len__(self) -> int:
        """Return the number of shots that have been run for this sequence."""

        return len(unwrap(self.session.sequences.get_shots(self.path)))

    def get_state(self) -> State:
        """Return the state of the sequence."""

        return unwrap(self.session.sequences.get_state(self.path))

    def get_global_parameters(self) -> ParameterNamespace:
        """Return a copy of the parameter tables set for this sequence.

        Raises:
            RuntimeError: If the sequence is in DRAFT state, since the global parameters
                are only set once the sequence has entered the PREPARING state.
        """

        return self.session.sequences.get_global_parameters(self.path)

    def get_iteration_configuration(self) -> IterationConfiguration:
        """Return the iteration configuration of the sequence."""

        return self.session.sequences.get_iteration_configuration(self.path)

    def get_time_lanes(self) -> TimeLanes:
        """Return the time lanes that define how a shot is run for this sequence."""

        return self.session.sequences.get_time_lanes(self.path)

    def set_time_lanes(self, time_lanes: TimeLanes) -> None:
        """Set the time lanes that define how a shot is run for this sequence."""

        return self.session.sequences.set_time_lanes(self.path, time_lanes)

    def get_shots(self) -> Iterable[Shot]:
        """Return the shots that belong to this sequence.

        The shots are returned sorted by index.
        """

        pure_shots = unwrap(self.session.sequences.get_shots(self.path))
        sorted_shots = sorted(pure_shots, key=lambda x: x.index)
        return (Shot.bound(shot, self.session) for shot in sorted_shots)

    def get_start_time(self) -> Optional[datetime.datetime]:
        """Return the time the sequence was started.

        If the sequence has not been started, return None.
        """

        return unwrap(self.session.sequences.get_stats(self.path)).start_time

    def get_end_time(self) -> Optional[datetime.datetime]:
        """Return the time the sequence was ended.

        If the sequence has not been ended, return None.
        """

        return unwrap(self.session.sequences.get_stats(self.path)).stop_time

    def get_expected_number_of_shots(self) -> int | Unknown:
        """Return the expected number of shots for the sequence.

        If the sequence has not been started, return None.
        """

        return unwrap(self.session.sequences.get_stats(self.path)).expected_number_shots

    def duplicate(self, target_path: PureSequencePath | str) -> Sequence:
        """Duplicate the sequence to a new path.

        The sequence created will be in the draft state and will have the same iteration
        configuration and time lanes as the original sequence.
        """

        if isinstance(target_path, str):
            target_path = PureSequencePath(target_path)

        iteration_configuration = self.get_iteration_configuration()
        time_lanes = self.get_time_lanes()
        self.create(target_path, iteration_configuration, time_lanes, self.session)
        return Sequence(target_path, self.session)

    def get_device_configurations(self) -> dict[DeviceName, DeviceConfiguration]:
        """Return the device configurations used when the sequence was launched."""

        device_configurations = self.session.sequences.get_device_configurations(
            self.path
        )

        return dict(device_configurations)

    def get_local_parameters(self) -> set[DottedVariableName]:
        """Return the name of the parameters specifically set for this sequence."""

        iterations = self.get_iteration_configuration()
        return iterations.get_parameter_names()

    def load_shots_data(
        self,
        importer: "DataImporter",
        tags: Optional[polars.type_aliases.FrameInitTypes] = None,
    ) -> Iterable[polars.DataFrame]:
        """Load the data of the shots that have been run for this sequence.

        Args:
            importer: A function that takes a shot and returns the data of the shot.
                It will be called for each shot in the sequence.
            tags: An optional object that can be converted to a 1-row DataFrame.
                The dataframe will be joined with the data of each shot.
                This allows to add extra information to the dataframes returned for this
                sequence.

        Yields:
            Dataframes containing the data of the shots in the sequence ordered by shot
            index.
        """

        if tags is not None:
            tags_dataframe = polars.DataFrame(tags)
            if len(tags_dataframe) != 1:
                raise ValueError("tags should be a single row DataFrame")
        else:
            tags_dataframe = None

        data_labels = set[DataLabel]()

        for shot in self.get_shots():
            # Here we use a trick to speed up data loading.
            # We keep track of the data labels that have been loaded so far and
            # reload them before passing them to the importer.
            # This ensure that data are queried in group, with is more efficient.
            shot.get_data_by_labels(data_labels)
            data = importer(shot)
            data_labels = set(shot._data_cache.keys())

            if tags is not None:
                yield data.join(tags_dataframe, how="cross")
            else:
                yield data



from typing import Mapping, Any, assert_never

import numpy as np

from caqtus.device import DeviceName, DeviceParameter
from caqtus.shot_compilation import (
    SequenceContext,
    DeviceNotUsedException,
    ShotContext,
)
from caqtus.shot_compilation.lane_compilers.timing import number_ticks, ns
from caqtus.types.recoverable_exceptions import InvalidValueError
from caqtus.types.timelane import CameraTimeLane, TakePicture
from ._configuration import CameraConfiguration
from ..sequencer import TimeStep
from ..sequencer.compilation import TriggerableDeviceCompiler
from ..sequencer.instructions import SequencerInstruction, Pattern, concatenate


class CameraCompiler(TriggerableDeviceCompiler):
    """Compiler for a camera device."""

    def __init__(self, device_name: DeviceName, sequence_context: SequenceContext):
        super().__init__(device_name, sequence_context)
        self.__device_name = device_name
        try:
            lane = sequence_context.get_lane(device_name)
        except KeyError:
            raise DeviceNotUsedException(device_name)
        if not isinstance(lane, CameraTimeLane):
            raise TypeError(
                f"Expected a camera time lane for device {device_name}, got "
                f"{type(lane)}"
            )
        self.__lane = lane
        configuration = sequence_context.get_device_configuration(device_name)
        if not isinstance(configuration, CameraConfiguration):
            raise TypeError(
                f"Expected a camera configuration for device {device_name}, got "
                f"{type(configuration)}"
            )
        self.__configuration = configuration

    def compile_initialization_parameters(self) -> Mapping[DeviceParameter, Any]:
        return {
            DeviceParameter("roi"): self.__configuration.roi,
            DeviceParameter("external_trigger"): True,
            DeviceParameter("timeout"): 1.0,
        }

    def compile_shot_parameters(self, shot_context: ShotContext) -> Mapping[str, Any]:
        step_durations = shot_context.get_step_durations()
        exposures = []
        picture_names = []
        shot_context.mark_lane_used(self.__device_name)
        for value, (start, stop) in zip(
            self.__lane.block_values(), self.__lane.block_bounds()
        ):
            if isinstance(value, TakePicture):
                exposure = sum(step_durations[start:stop])
                exposures.append(exposure)
                picture_names.append(value.picture_name)
        return {
            # Add a bit of extra time to the timeout, in case the shot takes a bit of
            # time to actually start.
            "timeout": shot_context.get_shot_duration() + 1,
            "picture_names": picture_names,
            "exposures": exposures,
        }

    def compute_trigger(
        self, sequencer_time_step: TimeStep, shot_context: ShotContext
    ) -> SequencerInstruction[np.bool_]:
        step_bounds = shot_context.get_step_start_times()

        instructions: list[SequencerInstruction[np.bool_]] = []
        for value, (start, stop) in zip(
            self.__lane.block_values(), self.__lane.block_bounds()
        ):
            length = number_ticks(
                step_bounds[start], step_bounds[stop], sequencer_time_step * ns
            )
            if isinstance(value, TakePicture):
                if length == 0:
                    raise InvalidValueError(
                        f"No trigger can be generated for picture "
                        f"'{value.picture_name}' because its exposure is too short"
                        f"({(step_bounds[stop] - step_bounds[start]) * 1e9} ns) with "
                        f"respect to the time step ({sequencer_time_step} ns)"
                    )
                instructions.append(Pattern([True]) * length)
            elif value is None:
                instructions.append(Pattern([False]) * length)
            else:
                assert_never(value)
        return concatenate(*instructions)



from enum import Enum
from typing import TypeAlias, TypeGuard

import attrs


class TriggerEdge(Enum):
    RISING = "rising"
    FALLING = "falling"
    BOTH = "both"

    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"


@attrs.define
class SoftwareTrigger:
    """Represents a trigger that is generated by a software command."""

    pass


@attrs.define
class ExternalTriggerStart:
    """A trigger generated by an external source once at the start of the sequence."""

    edge: TriggerEdge = TriggerEdge.RISING


@attrs.define
class ExternalClock:
    """A trigger generated by an external source for every time step."""

    edge: TriggerEdge = TriggerEdge.RISING


@attrs.define
class ExternalClockOnChange:
    """A trigger generated by an external source every time a value changes."""

    edge: TriggerEdge = TriggerEdge.RISING


Trigger: TypeAlias = (
    SoftwareTrigger | ExternalTriggerStart | ExternalClock | ExternalClockOnChange
)


def is_trigger(value) -> TypeGuard[Trigger]:
    return isinstance(
        value,
        (SoftwareTrigger, ExternalTriggerStart, ExternalClock, ExternalClockOnChange),
    )



from __future__ import annotations

import functools
from collections.abc import Mapping, Iterable
from typing import TypeAlias, TypeGuard, assert_never, Any

import attrs
import numpy

import caqtus.formatter as fmt
from caqtus.types.expression import Expression
from caqtus.types.parameter import (
    AnalogValue,
    is_analog_value,
    NotAnalogValueError,
    get_unit,
    add_unit,
    magnitude_in_unit,
)
from caqtus.utils import serialization
from .iteration_configuration import IterationConfiguration, Unknown
from ..recoverable_exceptions import EvaluationError
from ..units import DimensionalityError, InvalidDimensionalityError
from ..variable_name import DottedVariableName


def validate_step(instance, attribute, step):
    if is_step(step):
        return
    else:
        raise TypeError(f"Invalid step: {step}")


@attrs.define
class ContainsSubSteps:
    sub_steps: list[Step] = attrs.field(
        validator=attrs.validators.deep_iterable(
            iterable_validator=attrs.validators.instance_of(list),
            member_validator=validate_step,
        ),
        on_setattr=attrs.setters.validate,
    )


@attrs.define
class VariableDeclaration:
    """Represents the declaration of a variable.

    Attributes:
        variable: The name of the variable.
        value: The unevaluated to assign to the variable.
    """

    __match_args__ = ("variable", "value")

    variable: DottedVariableName = attrs.field(
        validator=attrs.validators.instance_of(DottedVariableName),
        on_setattr=attrs.setters.validate,
    )
    value: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self):
        return f"{self.variable} = {self.value}"


@attrs.define
class LinspaceLoop(ContainsSubSteps):
    """Represents a loop that iterates between two values with a fixed number of steps.

    Attributes:
        variable: The name of the variable that is being iterated over.
        start: The start value of the variable.
        stop: The stop value of the variable.
        num: The number of steps to take between the start and stop values.
    """

    __match_args__ = ("variable", "start", "stop", "num", "sub_steps")
    variable: DottedVariableName = attrs.field(
        validator=attrs.validators.instance_of(DottedVariableName),
        on_setattr=attrs.setters.validate,
    )
    start: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )
    stop: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )
    num: int = attrs.field(
        converter=int,
        validator=attrs.validators.ge(0),
        on_setattr=attrs.setters.pipe(attrs.setters.convert, attrs.setters.validate),
    )

    def __str__(self):
        return f"linspace loop over {fmt.shot_param(self.variable)}"

    def loop_values(
        self, evaluation_context: Mapping[DottedVariableName, Any]
    ) -> Iterable[AnalogValue]:
        """Returns the values that the variable represented by this loop takes.

        Args:
            evaluation_context: Contains the value of the variables with which to
                evaluate the start and stop expressions of the loop.

        Raises:
            EvaluationError: if the start or stop expressions could not be evaluated.
            NotAnalogValueError: if the start or stop expressions don't evaluate to an
                analog value.
            DimensionalityError: if the start or stop values are not commensurate.
        """

        start = self.start.evaluate(evaluation_context)
        if not is_analog_value(start):
            raise NotAnalogValueError(f"Start of {self} is not an analog value")
        stop = self.stop.evaluate(evaluation_context)
        if not is_analog_value(stop):
            raise NotAnalogValueError(f"Stop of {self} is not an analog value")

        unit = get_unit(start)

        start_magnitude = magnitude_in_unit(start, unit)
        try:
            stop_magnitude = magnitude_in_unit(stop, unit)
        except DimensionalityError as e:
            raise InvalidDimensionalityError(
                f"Start of {self} has invalid dimensionality"
            ) from e

        for value in numpy.linspace(start_magnitude, stop_magnitude, self.num):
            # val.item() is used to convert numpy scalar to python scalar
            value_with_unit = add_unit(value.item(), unit)
            yield value_with_unit


@attrs.define
class ArangeLoop(ContainsSubSteps):
    """Represents a loop that iterates between two values with a fixed step size.

    Attributes:
        variable: The name of the variable that is being iterated over.
        start: The start value of the variable.
        stop: The stop value of the variable.
        step: The step size between each value.
    """

    __match_args__ = ("variable", "start", "stop", "step", "sub_steps")
    variable: DottedVariableName = attrs.field(
        validator=attrs.validators.instance_of(DottedVariableName),
        on_setattr=attrs.setters.validate,
    )
    start: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )
    stop: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )
    step: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self):
        return f"arange loop over {fmt.shot_param(self.variable)}"

    def loop_values(
        self, evaluation_context: Mapping[DottedVariableName, Any]
    ) -> Iterable[AnalogValue]:
        """Returns the values that the variable represented by this loop takes.

        Args:
            evaluation_context: Contains the value of the variables with which to
                evaluate the start, stop and step expressions of the loop.

        Raises:
            EvaluationError: if the start, stop or step expressions could not be
                evaluated.
            NotAnalogValueError: if the start, stop or step expressions don't evaluate
                to an analog value.
            DimensionalityError: if the start, stop and step values are not
                commensurate.
        """

        start = self.start.evaluate(evaluation_context)
        if not is_analog_value(start):
            raise NotAnalogValueError(f"Start of {self} is not an analog value.")
        stop = self.stop.evaluate(evaluation_context)
        if not is_analog_value(stop):
            raise NotAnalogValueError(f"Stop of {self} is not an analog value.")
        step = self.step.evaluate(evaluation_context)
        if not is_analog_value(step):
            raise NotAnalogValueError(f"Step of {self} is not an analog value.")

        unit = get_unit(start)
        start_magnitude = magnitude_in_unit(start, unit)
        try:
            stop_magnitude = magnitude_in_unit(stop, unit)
        except DimensionalityError as e:
            raise InvalidDimensionalityError(
                f"Start of {self} has invalid dimensionality."
            ) from e
        try:
            step_magnitude = magnitude_in_unit(step, unit)
        except DimensionalityError as e:
            raise InvalidDimensionalityError(
                f"Step of {self} has invalid dimensionality."
            ) from e

        for value in numpy.arange(start_magnitude, stop_magnitude, step_magnitude):
            # val.item() is used to convert numpy scalar to python scalar
            value_with_unit = add_unit(value.item(), unit)
            yield value_with_unit


@attrs.define
class ExecuteShot:
    """Step that represents the execution of a shot."""

    pass


def unstructure_hook(execute_shot: ExecuteShot) -> str:
    return {"execute": "shot"}


def structure_hook(data: str, cls: type[ExecuteShot]) -> ExecuteShot:
    return ExecuteShot()


serialization.register_unstructure_hook(ExecuteShot, unstructure_hook)

serialization.register_structure_hook(ExecuteShot, structure_hook)

"""TypeAlias for the different types of steps."""
Step: TypeAlias = ExecuteShot | VariableDeclaration | LinspaceLoop | ArangeLoop


def is_step(step) -> TypeGuard[Step]:
    return isinstance(
        step,
        (
            ExecuteShot,
            VariableDeclaration,
            LinspaceLoop,
            ArangeLoop,
        ),
    )


@attrs.define
class StepsConfiguration(IterationConfiguration):
    """Define the parameter iteration of a sequence as a list of steps.

    Attributes:
        steps: The steps of the iteration.
    """

    steps: list[Step] = attrs.field(
        validator=attrs.validators.deep_iterable(
            iterable_validator=attrs.validators.instance_of(list),
            member_validator=validate_step,
        ),
        on_setattr=attrs.setters.validate,
    )

    def expected_number_shots(self) -> int | Unknown:
        return sum(expected_number_shots(step) for step in self.steps)

    def get_parameter_names(self) -> set[DottedVariableName]:
        return set().union(*[get_parameter_names(step) for step in self.steps])

    @classmethod
    def dump(cls, steps_configuration: StepsConfiguration) -> serialization.JSON:
        return serialization.unstructure(steps_configuration, StepsConfiguration)

    @classmethod
    def load(cls, data: serialization.JSON) -> StepsConfiguration:
        return serialization.structure(data, StepsConfiguration)


@functools.singledispatch
def expected_number_shots(step: Step) -> int | Unknown:  # type: ignore
    assert_never(step)


@expected_number_shots.register
def _(step: VariableDeclaration):
    return 0


@expected_number_shots.register
def _(step: ExecuteShot):
    return 1


@expected_number_shots.register
def _(step: LinspaceLoop):
    sub_steps_number = sum(
        expected_number_shots(sub_step) for sub_step in step.sub_steps
    )
    return sub_steps_number * step.num


@expected_number_shots.register
def _(step: ArangeLoop):
    try:
        length = len(list(step.loop_values({})))
    except (EvaluationError, NotAnalogValueError, DimensionalityError):
        # The errors above can occur if the steps are still being edited or if the
        # expressions depend on other variables that are not defined here.
        # These can be errors on the user side, so we don't want to crash on them, and
        # we just indicate that we don't know the number of shots.
        return Unknown()

    sub_steps_number = sum(
        expected_number_shots(sub_step) for sub_step in step.sub_steps
    )
    return sub_steps_number * length


def get_parameter_names(step: Step) -> set[DottedVariableName]:
    match step:
        case VariableDeclaration(variable=variable, value=_):
            return {variable}
        case ExecuteShot():
            return set()
        case LinspaceLoop(variable=variable, sub_steps=sub_steps) | ArangeLoop(
            variable=variable, sub_steps=sub_steps
        ):
            return {variable}.union(
                *[get_parameter_names(sub_step) for sub_step in sub_steps]
            )
        case _:
            assert_never(step)



import numpy as np

import caqtus.formatter as fmt
from caqtus.device.sequencer.instructions import (
    SequencerInstruction,
    Pattern,
    concatenate,
)
from caqtus.shot_compilation.compilation_contexts import ShotContext
from caqtus.shot_compilation.lane_compilers.timing import number_ticks, ns
from caqtus.types.expression import Expression
from caqtus.types.recoverable_exceptions import InvalidTypeError
from caqtus.types.timelane import DigitalTimeLane
from ..._time_step import TimeStep


def compile_digital_lane(
    lane: DigitalTimeLane,
    time_step: TimeStep,
    shot_context: ShotContext,
) -> SequencerInstruction[np.bool_]:
    step_names = shot_context.get_step_names()
    if len(lane) != len(step_names):
        raise ValueError(
            f"Number of steps in lane ({len(lane)}) does not match number of"
            f" steps ({len(step_names)})"
        )

    step_bounds = shot_context.get_step_start_times()
    instructions = []
    for cell_value, (start, stop) in zip(lane.block_values(), lane.block_bounds()):
        length = number_ticks(step_bounds[start], step_bounds[stop], time_step * ns)
        if isinstance(cell_value, bool):
            instructions.append(get_constant_instruction(cell_value, length))
        elif isinstance(cell_value, Expression):
            value = cell_value.evaluate(shot_context.get_variables())
            if not isinstance(value, bool):
                raise InvalidTypeError(
                    f"{fmt.expression(cell_value)} does not evaluate to "
                    f"{fmt.type_(bool)}, but to {fmt.type_(type(value))}",
                )
            instructions.append(get_constant_instruction(value, length))

        else:
            raise NotImplementedError(f"Unexpected value {cell_value} in digital lane")
    return concatenate(*instructions)


def get_constant_instruction(
    value: bool, length: int
) -> SequencerInstruction[np.bool_]:
    return Pattern([value]) * length


#
# elif isinstance(cell_value, Blink):
# period = (
#     cell_value.period.evaluate(variables | units).to("ns").magnitude
# )
# duty_cycle = (
#     Quantity(cell_value.duty_cycle.evaluate(variables | units))
#     .to(dimensionless)
#     .magnitude
# )
# if not 0 <= duty_cycle <= 1:
#     raise ShotEvaluationError(
#         f"Duty cycle '{cell_value.duty_cycle.body}' must be between 0 and"
#         f" 1, not {duty_cycle}"
#     )
# num_ticks_per_period, _ = divmod(period, time_step)
# num_ticks_high = math.ceil(num_ticks_per_period * duty_cycle)
# num_ticks_low = num_ticks_per_period - num_ticks_high
# num_clock_pulses, remainder = divmod(length, num_ticks_per_period)
# phase = (
#     Quantity(cell_value.phase.evaluate(variables | units))
#     .to(dimensionless)
#     .magnitude
# )
# if not 0 <= phase <= 2 * math.pi:
#     raise ShotEvaluationError(
#         f"Phase '{cell_value.phase.body}' must be between 0 and 2*pi, not"
#         f" {phase}"
#     )
# split_position = round(phase / (2 * math.pi) * num_ticks_per_period)
# clock_pattern = (
#         Pattern([True]) * num_ticks_high + Pattern([False]) * num_ticks_low
# )
# a, b = clock_pattern[:split_position], clock_pattern[split_position:]
# clock_pattern = b + a
# pattern = (
#         clock_pattern * num_clock_pulses + Pattern([False]) * remainder
# )
# if not len(pattern) == length:
#     raise RuntimeError(
#         f"Pattern length {len(pattern)} does not match expected length"
#         f" {length}"
#     )
# print(f"{pattern=}")



from ._proxy import SequencerProxy, SequenceStatusProxy
from .instructions import SequencerInstruction
from .trigger import SoftwareTrigger
from .._controller import DeviceController


class SequencerController(DeviceController):
    """Controls a sequencer during a shot."""

    async def run_shot(
        self,
        sequencer: SequencerProxy,
        /,
        sequence: SequencerInstruction,
        *args,
        **kwargs,
    ) -> None:
        trigger = await sequencer.get_trigger()
        async with sequencer.program_sequence(sequence) as programmed_sequence:
            if isinstance(trigger, SoftwareTrigger):
                await self.wait_all_devices_ready()
                async with programmed_sequence.run() as sequence_status:
                    await self.wait_until_finished(sequence_status)
            else:
                async with programmed_sequence.run() as sequence_status:
                    await self.wait_all_devices_ready()
                    await self.wait_until_finished(sequence_status)

    async def wait_until_finished(self, status: SequenceStatusProxy) -> None:
        while not await status.is_finished():
            await self.sleep(0)



from __future__ import annotations

import abc
import contextlib
import functools
import logging
import math
from collections.abc import Callable
from typing import (
    TypeVar,
    ParamSpec,
    TYPE_CHECKING,
    final,
    Optional,
    TypedDict,
)

import anyio
import anyio.to_thread

import caqtus.formatter as fmt
from caqtus.types.data import DataLabel, Data
from ._name import DeviceName
from .remote import DeviceProxy
from ..types.recoverable_exceptions import RecoverableException

if TYPE_CHECKING:
    from caqtus.experiment_control._shot_handling import ShotEventDispatcher
    from caqtus.experiment_control.shot_timing import ShotTimer

logger = logging.getLogger(__name__)


_T = TypeVar("_T")
_Q = ParamSpec("_Q")


class DeviceController[DeviceProxyType: DeviceProxy, **_P](abc.ABC):
    """Controls a device during a shot.

    Subclasses must implement :meth:`run_shot` to define the behavior of the associated
    device during a shot.

    This class is generic in the type of device proxy that it controls and the extra
    arguments passed to the :meth:`run_shot` method.

    Attributes:
        device_name: The name of the device being controlled.
    """

    def __init__(
        self,
        device_name: DeviceName,
        shot_event_dispatcher: "ShotEventDispatcher",
    ):
        # This method is not meant to be overridden or called directly.

        self.device_name = device_name
        self._event_dispatcher = shot_event_dispatcher
        self._signaled_ready = anyio.Event()
        self._signaled_ready_time: Optional[float] = None
        self._finished_waiting_ready_time: Optional[float] = None
        self._thread_times: list[tuple[str, float, float]] = []
        self._data_waits: list[tuple[str, float, float]] = []
        self._data_signals: list[tuple[str, float]] = []

    @abc.abstractmethod
    async def run_shot(
        self, device: DeviceProxyType, /, *args: _P.args, **kwargs: _P.kwargs
    ) -> None:
        """Runs a shot on the device.

        This method must call :meth:`wait_all_devices_ready` exactly once.

        Args:
            device: An asynchronous proxy to the device being controlled.
            args, kwargs: Extra arguments than can be computed before the shot
                and that are required to run the shot.
        """

        raise NotImplementedError

    @final
    async def _run_shot(
        self,
        device: DeviceProxyType,
        timeout: float,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> ShotStats:
        start_time = self._event_dispatcher.shot_time()
        try:
            with fail_after(delay=timeout, shield=False):
                await self.run_shot(device, *args, **kwargs)

            finished_time = self._event_dispatcher.shot_time()
            if not self._signaled_ready.is_set():
                raise RuntimeError(f"wait_all_devices_ready was not called in run_shot")
        except BaseException as e:
            raise RuntimeError(
                f"Error occurred for {fmt.device(self.device_name)}"
            ) from e
        assert self._signaled_ready_time is not None
        assert self._finished_waiting_ready_time is not None

        return ShotStats(
            start_time=start_time,
            signaled_ready_time=self._signaled_ready_time,
            finished_waiting_ready_time=self._finished_waiting_ready_time,
            finished_time=finished_time,
            thread_stats=self._thread_times,
            data_waits=self._data_waits,
            data_signals=self._data_signals,
        )

    @final
    async def wait_all_devices_ready(self) -> ShotTimer:
        """Wait for all devices to be ready for time-sensitive operations.

        This method must be called once the device has been programmed for the shot and
        is ready to be triggered or to react to data acquisition signals.

        It must be called exactly once in :meth:`run_shot`.

        The method will wait for all devices to be ready before returning.

        Returns:
            A timer with its start time set to the time when this function returns.
        """

        if self._signaled_ready.is_set():
            raise RuntimeError(
                f"wait_all_devices_ready must be called exactly once for {self}"
            )
        self._signaled_ready.set()
        self._signaled_ready_time = self._event_dispatcher.shot_time()
        timer = await self._event_dispatcher.wait_all_devices_ready()
        self._finished_waiting_ready_time = self._event_dispatcher.shot_time()
        return timer

    @final
    def signal_data_acquired(self, label: DataLabel, data: Data) -> None:
        """Signals that data has been acquired from the device.

        Args:
            label: The label of the data.
            data: The data that was acquired.
        """

        self._event_dispatcher.signal_data_acquired(self.device_name, label, data)
        self._data_signals.append((label, self._event_dispatcher.shot_time()))

    @final
    async def wait_data_acquired(self, label: DataLabel) -> Data:
        """Waits until another device signals that some data has been acquired."""

        start = self._event_dispatcher.shot_time()
        data = await self._event_dispatcher.wait_data_acquired(self.device_name, label)
        end = self._event_dispatcher.shot_time()
        self._data_waits.append((label, start, end))
        return data

    def _debug_stats(self):
        return {
            "signaled_ready_time": self._signaled_ready_time,
            "finished_waiting_ready_time": self._finished_waiting_ready_time,
        }

    @final
    async def run_in_thread(
        self, func: Callable[_Q, _T], *args: _Q.args, **kwargs: _Q.kwargs
    ) -> _T:
        func_name = func.__name__
        start_time = self._event_dispatcher.shot_time()
        result = await anyio.to_thread.run_sync(
            functools.partial(func, *args, **kwargs)
        )
        end_time = self._event_dispatcher.shot_time()
        self._thread_times.append((func_name, start_time, end_time))
        return result

    @final
    async def sleep(self, seconds: float) -> None:
        """Sleeps for a given number of seconds."""

        await anyio.sleep(seconds)


class ShotStats(TypedDict):
    start_time: float
    signaled_ready_time: float
    finished_waiting_ready_time: float
    finished_time: float
    thread_stats: list[tuple[str, float, float]]
    data_waits: list[tuple[str, float, float]]
    data_signals: list[tuple[str, float]]


DeviceControllerType = TypeVar("DeviceControllerType", bound=DeviceController)


@contextlib.contextmanager
def fail_after(delay: float | None, shield: bool = False):
    """A copy of anyio.fail_after but with custom timeout exception."""

    deadline = (anyio.current_time() + delay) if delay is not None else math.inf
    with anyio.CancelScope(deadline=deadline, shield=shield) as cancel_scope:
        yield cancel_scope

    if cancel_scope.cancelled_caught and anyio.current_time() >= cancel_scope.deadline:
        raise DeviceTimeoutError(f"Timeout after {delay:.2f} seconds")


class DeviceTimeoutError(TimeoutError, RecoverableException):
    """Raise when a device exceeded the shot timeout."""

    pass



"""Contains utility functions that are not-specific to the package functionalities."""

from ._add_exc_note import add_exc_note
from ._log_duration import log_duration
from ._log_exception import log_exception
from .duration_timer import DurationTimer, DurationTimerLog

__all__ = [
    "log_exception",
    "log_duration",
    "DurationTimer",
    "DurationTimerLog",
    "add_exc_note",
]



#!/usr/bin/python
from PySide6 import QtWidgets, QtCore

from ...widgets.dialogs import FileDialog
from .prop_widgets_abstract import BaseProperty


class PropFilePath(BaseProperty):
    """
    Displays a node property as a "QFileDialog" open widget in the
    PropertiesBin.
    """

    def __init__(self, parent=None):
        super(PropFilePath, self).__init__(parent)
        self._ledit = QtWidgets.QLineEdit()
        self._ledit.setAlignment(QtCore.Qt.AlignLeft)
        self._ledit.editingFinished.connect(self._on_value_change)
        self._ledit.clearFocus()

        icon = self.style().standardIcon(QtWidgets.QStyle.StandardPixmap(21))
        _button = QtWidgets.QPushButton()
        _button.setIcon(icon)
        _button.clicked.connect(self._on_select_file)

        hbox = QtWidgets.QHBoxLayout(self)
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.addWidget(self._ledit)
        hbox.addWidget(_button)

        self._ext = "*"
        self._file_directory = None

    def _on_select_file(self):
        file_path = FileDialog.getOpenFileName(
            self, file_dir=self._file_directory, ext_filter=self._ext
        )
        file = file_path[0] or None
        if file:
            self.set_value(file)

    def _on_value_change(self, value=None):
        if value is None:
            value = self._ledit.text()
        self.set_file_directory(value)
        self.value_changed.emit(self.get_name(), value)

    def set_file_ext(self, ext=None):
        self._ext = ext or "*"

    def set_file_directory(self, directory):
        self._file_directory = directory

    def get_value(self):
        return self._ledit.text()

    def set_value(self, value):
        _value = str(value)
        if _value != self.get_value():
            self._ledit.setText(_value)
            self._on_value_change(_value)


class PropFileSavePath(PropFilePath):
    """
    Displays a node property as a "QFileDialog" save widget in the
    PropertiesBin.
    """

    def _on_select_file(self):
        file_path = FileDialog.getSaveFileName(
            self, file_dir=self._file_directory, ext_filter=self._ext
        )
        file = file_path[0] or None
        if file:
            self.set_value(file)



from collections.abc import Sequence
from typing import Optional

import numpy as np
from PySide6.QtCharts import (
    QChart,
    QLineSeries,
    QChartView,
    QValueAxis,
)
from PySide6.QtCore import (
    QAbstractTableModel,
    Qt,
    QSortFilterProxyModel,
    QModelIndex,
    QPointF,
)
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import (
    QWidget,
    QApplication,
    QItemEditorFactory,
    QDoubleSpinBox,
    QStyledItemDelegate,
)

from caqtus.gui._common.NodeGraphQt import BaseNode, NodeBaseWidget
from caqtus.gui.condetrol._icons import get_icon
from caqtus.types.parameter import add_unit, magnitude_in_unit
from caqtus.types.units import Unit
from caqtus.types.units.base import convert_to_base_units
from caqtus.utils.itertools import pairwise
from .calibrated_analog_mapping_widget_ui import Ui_CalibratedAnalogMappingWigdet


class CalibratedAnalogMappingNode(BaseNode):
    __identifier__ = "caqtus.sequencer_node.mapping"
    NODE_NAME = "Analog mapping"

    def __init__(self):
        super().__init__()
        self.add_output("out", multi_output=False, display_name=False)
        self.input_port = self.add_input("in", multi_input=False)
        self._widget = NodeWidgetWrapper(self.view)
        self.add_custom_widget(self._widget)

    def set_units(
        self, input_units: Optional[str], output_units: Optional[str]
    ) -> None:
        self._widget._widget.set_units(input_units, output_units)

    def set_data_points(self, values: Sequence[tuple[float, float]]) -> None:
        self._widget._widget.set_data_points(values)

    def get_input_node(self) -> Optional[BaseNode]:
        input_nodes = self.connected_input_nodes()[self.input_port]
        if len(input_nodes) == 0:
            return None
        elif len(input_nodes) == 1:
            return input_nodes[0]
        else:
            assert False, "There can't be multiple nodes connected to the input"

    def get_units(self) -> tuple[Optional[str], Optional[str]]:
        return self._widget._widget.get_units()

    def get_data_points(self) -> list[tuple[float, float]]:
        return self._widget._widget.get_data_points()


class NodeWidgetWrapper(NodeBaseWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        # set the label above the widget.
        self.set_label("Custom Widget")

        # set the custom widget.
        self._widget = CalibratedAnalogMappingWidget()
        self._widget.setStyle(QApplication.style())
        self.set_custom_widget(self._widget)

    def get_value(self):
        return ""

    def set_value(self, text):
        pass


class CalibratedAnalogMappingWidget(QWidget, Ui_CalibratedAnalogMappingWigdet):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setupUi(self)

        self._model = Model(self)
        self._sorted_model = QSortFilterProxyModel(self)
        self._sorted_model.setSourceModel(self._model)
        self._sorted_model.sort(0, Qt.SortOrder.AscendingOrder)
        self.tableView.setModel(self._sorted_model)
        self.add_button.setIcon(get_icon("plus"))
        self.remove_button.setIcon(get_icon("minus"))
        self.add_button.clicked.connect(self.on_add_button_clicked)
        self.remove_button.clicked.connect(self.on_remove_button_clicked)

        delegate = QStyledItemDelegate(self)
        delegate.setItemEditorFactory(ItemEditorFactory())
        self.tableView.setItemDelegate(delegate)

        self._chart = QChart()
        self._chart.setAnimationOptions(QChart.AnimationOption.AllAnimations)
        self._series = QLineSeries()
        self._series.setName("Values")

        self._model.dataChanged.connect(self._update_series)
        self._model.rowsInserted.connect(self._update_series)
        self._model.rowsRemoved.connect(self._update_series)
        self._model.modelReset.connect(self._update_series)
        self._chart.addSeries(self._series)

        self.x_axis = QValueAxis()
        self.x_axis.setTitleText("Input")
        self._chart.addAxis(self.x_axis, Qt.AlignmentFlag.AlignBottom)
        self._series.attachAxis(self.x_axis)

        self.y_axis = QValueAxis()
        self.y_axis.setTitleText("Output")
        self._chart.addAxis(self.y_axis, Qt.AlignmentFlag.AlignRight)
        self._series.attachAxis(self.y_axis)

        self._chart.layout().setContentsMargins(0, 0, 0, 0)
        self._chartView = QChartView(self._chart, self)
        self._chartView.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.inputUnitLineEdit.textChanged.connect(self.set_input_units)
        self.outputUnitLineEdit.textChanged.connect(self.set_output_units)

        self.tabWidget.insertTab(0, self._chartView, "Curve")
        self.tabWidget.setCurrentIndex(0)

    def _update_series(self):
        self._series.clear()
        number_points = self._sorted_model.rowCount()
        if number_points == 0:
            return
        x_points = []
        y_points = []
        for row in range(number_points):
            x_points.append(self._sorted_model.data(self._sorted_model.index(row, 0)))
            y_points.append(self._sorted_model.data(self._sorted_model.index(row, 1)))
        input_units = Unit(u) if (u := self.input_units()) else None
        input_base_points, input_base_units = convert_to_base_units(
            np.array(x_points), input_units
        )

        output_units = Unit(u) if (u := self.output_units()) else None
        y_base_points, y_base_units = convert_to_base_units(
            np.array(y_points), output_units
        )

        input_points = np.concat([np.linspace(a, b, 50) for a, b in pairwise(x_points)])
        interpolation_input_base_points, _ = convert_to_base_units(
            input_points, input_units
        )

        output_points_base = np.interp(
            interpolation_input_base_points, input_base_points, y_base_points
        )
        output_points = magnitude_in_unit(
            add_unit(output_points_base, y_base_units), output_units
        )

        new_points = [QPointF(x, y) for x, y in zip(input_points, output_points)]
        self._series.append(new_points)
        self.auto_scale()

    def set_data_points(self, values: Sequence[tuple[float, float]]) -> None:
        self._model.set_values(values)

    def set_input_units(self, input_units: Optional[str]) -> None:
        if input_units:
            self.inputUnitLineEdit.setText(input_units)
            self.x_axis.setTitleText(f"Input [{input_units}]")
        else:
            # This handles both the case where the input_units is None and the case
            # where the input_units is an empty string.
            self.inputUnitLineEdit.clear()
            self.x_axis.setTitleText("Input")
        self._update_series()

    def set_output_units(self, output_units: Optional[str]) -> None:
        if output_units:
            self.outputUnitLineEdit.setText(output_units)
            self.y_axis.setTitleText(f"Output [{output_units}]")
        else:
            # This handles both the case where the input_units is None and the case
            # where the input_units is an empty string.
            self.outputUnitLineEdit.clear()
            self.y_axis.setTitleText("Output")
        self._update_series()

    def set_units(
        self, input_units: Optional[str], output_units: Optional[str]
    ) -> None:
        self.set_input_units(input_units)
        self.set_output_units(output_units)

    def get_data_points(self) -> list[tuple[float, float]]:
        return self._model.get_values()

    def get_units(self) -> tuple[Optional[str], Optional[str]]:
        return self.input_units(), self.output_units()

    def input_units(self) -> Optional[str]:
        input_units = self.inputUnitLineEdit.text()
        if input_units == "":
            return None
        return input_units

    def output_units(self) -> Optional[str]:
        output_units = self.outputUnitLineEdit.text()
        if output_units == "":
            return None
        return output_units

    def auto_scale(self) -> None:
        self._chart.axisX().setRange(*self._model.x_range())
        self._chart.axisY().setRange(*self._model.y_range())

    def on_add_button_clicked(self):
        self._model.insertRow(
            self._sorted_model.mapToSource(self.tableView.currentIndex()).row() + 1
        )

    def on_remove_button_clicked(self):

        self._model.removeRow(
            self._sorted_model.mapToSource(self.tableView.currentIndex()).row()
        )


class Model(QAbstractTableModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._values = []

    def rowCount(self, parent):
        return len(self._values)

    def columnCount(self, parent):
        return 2

    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            if index.column() == 0:
                return self._values[index.row()][0]
            elif index.column() == 1:
                return self._values[index.row()][1]
        return None

    def setData(self, index, value, role):
        if role == Qt.ItemDataRole.EditRole:
            if index.column() == 0:
                self._values[index.row()] = (value, self._values[index.row()][1])
            elif index.column() == 1:
                self._values[index.row()] = (self._values[index.row()][0], value)
            self.dataChanged.emit(index, index)
            return True
        return False

    def set_values(self, values: Sequence[tuple[float, float]]):
        self.beginResetModel()
        self._values = list(values)
        self.endResetModel()

    def get_values(self) -> list[tuple[float, float]]:
        return sorted(self._values)

    def x_range(self):
        return min(x for x, _ in self._values), max(x for x, _ in self._values)

    def y_range(self):
        return min(y for _, y in self._values), max(y for _, y in self._values)

    def flags(self, index):
        return (
            Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
        )

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                if section == 0:
                    return "Input"
                elif section == 1:
                    return "Output"
            else:
                return str(section)
        return None

    def insertRow(self, row, parent=QModelIndex()):
        self.beginInsertRows(parent, row, row)
        self._values.insert(row, (0.0, 0.0))
        self.endInsertRows()
        return True

    def removeRow(self, row, parent=QModelIndex()):
        self.beginRemoveRows(parent, row, row)
        del self._values[row]
        self.endRemoveRows()
        return True


class ItemEditorFactory(QItemEditorFactory):
    def __init__(self):
        super().__init__()

    def createEditor(self, userType, parent):
        if userType == 6:
            spin_box = QDoubleSpinBox(parent)
            spin_box.setDecimals(3)
            spin_box.setMaximum(1000)
            spin_box.setMinimum(-1000)
            return spin_box
        else:
            return super().createEditor(userType, parent)



"""This module contains the implementation of timed instructions."""

from ._instructions import (
    SequencerInstruction,
    Concatenated,
    Repeated,
    Pattern,
    concatenate,
)
from ._plot import plot_instruction
from ._ramp import create_ramp, Ramp
from ._stack import stack_instructions, merge_instructions
from ._to_graph import to_graph
from ._to_time_array import convert_to_change_arrays
from ._with_name import with_name

__all__ = [
    "SequencerInstruction",
    "Concatenated",
    "Repeated",
    "Pattern",
    "convert_to_change_arrays",
    "with_name",
    "stack_instructions",
    "merge_instructions",
    "concatenate",
    "create_ramp",
    "Ramp",
    "plot_instruction",
    "to_graph",
]



import trio

from caqtus.utils._no_public_constructor import NoPublicConstructor


class ShotTimer(metaclass=NoPublicConstructor):
    """Gives access to pseudo-real time primitives during a shot.

    It gives the possibility to react at specific times during a shot.

    All times are relative to the start of the shot and are in seconds.
    """

    def __init__(self) -> None:
        # This class relies on the trio implementation.
        # It would need to be adapted for anyio/asyncio.
        self._start_time = trio.current_time()

    def elapsed(self) -> float:
        """Returns the elapsed time since the start of the shot."""

        return trio.current_time() - self._start_time

    async def wait_until(self, target_time: float) -> float:
        """Waits until a target time is reached.

        Args:
            target_time: The target time relative to the start of the shot.
                The target time can be in the past, in which case the function will
                return immediately.

        Returns:
            The duration waited for the target time to be reached.
            This duration can be negative if the target time is in the past.

        Raises:
            ValueError: If the target time is negative.

        Warning:
            This function is not guaranteed to be precise.
            Its accuracy depends on the underlying operating system and the event loop
            load.
        """

        if target_time < 0:
            raise ValueError("The target time must be positive.")

        duration_to_sleep = target_time - self.elapsed()

        if duration_to_sleep < 0:
            # We still need to await to get a checkpoint.
            await trio.sleep(0)
        else:
            await trio.sleep(duration_to_sleep)
        return duration_to_sleep



from typing import Protocol

from caqtus.device import DeviceConfiguration
from caqtus.utils.serialization import JSON


class DeviceConfigJSONSerializerProtocol(Protocol):
    def dump_device_configuration(
        self, config: DeviceConfiguration
    ) -> tuple[str, JSON]: ...

    def load_device_configuration(
        self, tag: str, content: JSON
    ) -> DeviceConfiguration: ...



from typing import TypeGuard, Any, TypeAlias

from caqtus.utils import serialization
from .analog_value import AnalogValue, is_analog_value, Quantity

Parameter: TypeAlias = AnalogValue | int | bool


def unstructure_quantity(value: Quantity):
    return float(value.magnitude), f"{value.units:~}"


def structure_quantity(value: Any, _) -> Quantity:
    try:
        return Quantity(*value)
    except TypeError:
        raise ValueError(f"Cannot structure {value!r} as a Quantity.")


serialization.register_unstructure_hook(Quantity, unstructure_quantity)

serialization.register_structure_hook(Quantity, structure_quantity)


def is_parameter(parameter: Any) -> TypeGuard[Parameter]:
    """Returns True if the value is a valid parameter type, False otherwise."""

    return is_analog_value(parameter) or isinstance(parameter, (int, bool))



from typing import Generic, TypeVar

T = TypeVar("T")


class Proxy(Generic[T]):
    def __init__(self, pid: int, obj_id: int):
        self._pid = pid
        self._obj_id = obj_id

    def __repr__(self):
        return f"Proxy(pid={self._pid}, obj_id={self._obj_id})"



from __future__ import annotations

import abc
from collections.abc import Mapping
from typing import Any, TypeAlias, Union

import attrs

import caqtus.formatter as fmt
from caqtus.types.expression import Expression
from caqtus.types.parameter import (
    Parameter,
    is_parameter,
    is_quantity,
)
from caqtus.types.recoverable_exceptions import InvalidTypeError
from caqtus.types.variable_name import DottedVariableName


@attrs.define
class Transformation(abc.ABC):
    """Defines a transformation that can be applied to produce an output value."""

    def evaluate(self, variables: Mapping[DottedVariableName, Any]) -> OutputValue:
        """Evaluates the transformation using the given variables."""

        raise NotImplementedError


OutputValue: TypeAlias = Parameter
"""A value that can be used to compute the output of a device."""

EvaluableOutput: TypeAlias = Union[Expression, Transformation]
"""Defines an operation that can be evaluated to an output value.

Evaluable object can be used in the :func:`evaluate` function.
"""


def evaluate(
    input_: EvaluableOutput, variables: Mapping[DottedVariableName, Any]
) -> OutputValue:
    """Evaluates the input and returns the result as a parameter."""

    evaluated = input_.evaluate(variables)

    if not is_parameter(evaluated):
        if isinstance(input_, Expression):
            raise InvalidTypeError(
                f"{fmt.expression(input_)} does not evaluate to a parameter, "
                f"got {fmt.type_(type(evaluated))}.",
            )
        # We only make the error recoverable if the user messed up the expression.
        # Otherwise, it is a bug because transformations should always return an output
        # value.
        raise TypeError(f"Expected {OutputValue}, got {type(evaluated)}.")

    if is_quantity(evaluated):
        # We convert to base units ASAP to avoid dealing with non-linear units like dB
        # and dBm.
        return evaluated.to_base_units()
    return evaluated


evaluable_output_validator = attrs.validators.instance_of((Expression, Transformation))



from __future__ import annotations

import abc
import datetime
import functools
from collections.abc import Callable
from typing import Optional, Mapping, assert_never, TypeAlias

import anyio
import attrs
from PySide6.QtCore import QSettings, Qt, QByteArray
from PySide6.QtWidgets import (
    QWidget,
    QMainWindow,
    QFileDialog,
    QMdiArea,
    QDockWidget,
)

from caqtus.gui._common.sequence_hierarchy import AsyncPathHierarchyView
from caqtus.session import (
    ExperimentSessionMaker,
    PureSequencePath,
    AsyncExperimentSession,
)
from caqtus.session import PathNotFoundError, PathIsNotSequenceError, Shot
from caqtus.session._return_or_raise import unwrap
from caqtus.session._sequence_collection import PureShot
from caqtus.utils import serialization
from caqtus.utils.serialization import JSON
from .main_window_ui import Ui_ShotViewerMainWindow
from .workspace import ViewState, WorkSpace
from ..single_shot_viewers import ShotView, ManagerName

ViewCreator: TypeAlias = Callable[[], tuple[str, ShotView]]


class SnapShotWindowHandler:
    def __init__(
        self,
        window: SnapShotMainWindow,
        experiment_session_maker: ExperimentSessionMaker,
    ):
        super().__init__()
        self.window = window
        self.experiment_session_maker = experiment_session_maker
        self.task_group = anyio.create_task_group()
        self.state_lock = anyio.Lock()

        self._state: WidgetState = NoSequenceSelected()
        # Lock to prevent changing the state while comparing if the state is different
        # from the stored sequence.
        self._state_lock = anyio.Lock()

        self.window.hierarchy_view.sequence_double_clicked.connect(
            self.on_sequence_double_clicked
        )

    async def exec_async(self):
        async with self.task_group:
            self.task_group.start_soon(self.watch)
            self.task_group.start_soon(self.window.hierarchy_view.run_async)

    async def watch(self):
        while True:
            await self.update_state()
            await anyio.sleep(50e-3)

    def on_sequence_double_clicked(self, path: PureSequencePath) -> None:
        # We can't transition synchronously now, because the widget might be in the
        # middle of a state comparison or a transition.
        # So instead we schedule the transition to happen asynchronously once it's safe.
        self.task_group.start_soon(self._on_sequence_double_clicked_async, path)

    async def _on_sequence_double_clicked_async(self, path: PureSequencePath) -> None:
        async with (
            self._state_lock,
            self.experiment_session_maker.async_session() as session,
        ):
            state = await get_state_async(path, session)
            await self._transition(state)

    async def update_state(self) -> None:
        """Ensure that the widget displays the up-to-date state of the sequence."""

        async with self._state_lock:
            match self._state:
                case NoSequenceSelected():
                    return
                case SequenceSelected(path=path):
                    async with (
                        self.experiment_session_maker.async_session() as session,
                    ):
                        new_state = await get_state_async(path, session)
                        if new_state != self._state:
                            await self._transition(new_state)
                case _:
                    assert_never(self._state)

    async def _transition(self, state: WidgetState) -> None:
        if isinstance(state, SequenceSelected):
            if state.shots:
                last_shot = max(state.shots, key=lambda s: s.index)
                await self._update_views(last_shot)
        self._state = state

    async def _update_views(self, shot: PureShot) -> None:
        async with anyio.create_task_group() as tg:
            for view in self.window.get_views().values():
                tg.start_soon(self._update_view, shot, view)

    async def _update_view(self, shot: PureShot, view: ShotView) -> None:
        # We need to create a new session for each view, because otherwise the views
        # might use the same session concurrently, which is not allowed.
        with self.experiment_session_maker() as session:
            bound_shot = Shot.bound(shot, session)
            await view.display_shot(bound_shot)


class SnapShotMainWindow(QMainWindow, Ui_ShotViewerMainWindow):
    """The main window of the shot viewer application.

    This window displays the sequences in a dock widget.
    The central widget is a multi-document interface area where the user can open
    multiple views of the shots.
    When a sequence is double-clicked in the sequence dock widget, the window will
    display the last shot of the sequence in all the views.
    """

    def __init__(
        self,
        hierarchy_view: AsyncPathHierarchyView,
        view_creators: Mapping[str, ViewCreator],
        view_dumper: Callable[[ShotView], JSON],
        view_loader: Callable[[JSON], ShotView],
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)

        self._views: dict[str, tuple[ManagerName, ShotView]] = {}
        self._view_creators = view_creators
        self._mdi_area = QMdiArea()

        self.hierarchy_view = hierarchy_view
        self._setup_ui()
        self.restore_state()

        self._view_dumper = view_dumper
        self._view_loader = view_loader

        self.tile_action.triggered.connect(self._mdi_area.tileSubWindows)

    def restore_state(self):
        ui_settings = QSettings("Caqtus", "ShotViewer")
        self.restoreState(ui_settings.value("state", self.saveState()))
        self.restoreGeometry(ui_settings.value("geometry", self.saveGeometry()))

    def closeEvent(self, a0):
        ui_settings = QSettings("Caqtus", "ShotViewer")
        ui_settings.setValue("state", self.saveState())
        ui_settings.setValue("geometry", self.saveGeometry())
        super().closeEvent(a0)

    def _setup_ui(self) -> None:
        self.setupUi(self)
        paths_dock = QDockWidget("Sequences", self)
        paths_dock.setObjectName("SequencesDock")
        paths_dock.setWidget(self.hierarchy_view)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, paths_dock)

        self.action_save_workspace_as.triggered.connect(self.save_workspace_as)
        self.action_load_workspace.triggered.connect(self.load_workspace)

        self._add_view_creators(self._view_creators)

        self.setWindowTitle("Single Shot Viewer")
        self.setCentralWidget(self._mdi_area)

    def _add_view_creators(self, view_creators: Mapping[str, ViewCreator]) -> None:
        for name in view_creators:
            self.menu_add_view.addAction(name).triggered.connect(
                functools.partial(self._create_view, name)
            )

    def _create_view(self, creator_name: str) -> None:
        manager = self._view_creators[creator_name]
        name, view = manager()
        self._add_view(name, view)

    def _add_view(self, view_name: str, view: ShotView) -> None:
        sub_window = self._mdi_area.addSubWindow(view)
        sub_window.setWindowTitle(view_name)
        sub_window.show()

    def get_views(self) -> dict[str, ShotView]:
        return {
            sub_window.windowTitle(): sub_window.widget()
            for sub_window in self._mdi_area.subWindowList()
        }

    def get_workspace(self) -> WorkSpace:
        view_states = {}
        for sub_window in self._mdi_area.subWindowList():
            view_name = sub_window.windowTitle()
            view = sub_window.widget()
            assert isinstance(view, ShotView)
            view_states[view_name] = ViewState(
                view_state=self._view_dumper(view),
                window_geometry=_bytes_to_str(sub_window.saveGeometry()),
            )
        window_state = _bytes_to_str(self.saveState())
        window_geometry = _bytes_to_str(self.saveGeometry())
        return WorkSpace(
            view_states=view_states,
            window_state=window_state,
            window_geometry=window_geometry,
        )

    def set_workspace(self, workspace: WorkSpace) -> None:
        self.clear()

        self.restoreState(_str_to_bytes_array(workspace.window_state))
        self.restoreGeometry(_str_to_bytes_array(workspace.window_geometry))

        for view_name, view_state in workspace.view_states.items():
            view = self._view_loader(view_state.view_state)
            sub_window = self._mdi_area.addSubWindow(view)
            sub_window.setWindowTitle(view_name)
            sub_window.restoreGeometry(_str_to_bytes_array(view_state.window_geometry))
            sub_window.show()

    def clear(self):
        for sub_window in self._mdi_area.subWindowList():
            sub_window.close()

    def save_workspace_as(self) -> None:
        file_name, _ = QFileDialog.getSaveFileName(
            self,
            "Save current workspace",
            "",
            "JSON (*.json)",
        )
        if file_name:
            workspace = self.get_workspace()
            json_string = serialization.to_json(workspace)
            with open(file_name, "w") as f:
                f.write(json_string)

    def load_workspace(self) -> None:
        file_name, _ = QFileDialog.getOpenFileName(
            self,
            "Load workspace",
            "",
            "JSON (*.json)",
        )
        if file_name:
            with open(file_name, "r") as f:
                json_string = f.read()
            workspace = serialization.from_json(json_string, WorkSpace)
            self.set_workspace(workspace)


def _str_to_bytes_array(string: str) -> QByteArray:
    return QByteArray.fromHex(bytes(string, "ascii"))


def _bytes_to_str(array: QByteArray) -> str:
    return bytes(array.toHex()).decode("ascii")


@attrs.frozen
class WidgetState(abc.ABC):
    pass


@attrs.frozen
class NoSequenceSelected(WidgetState):
    pass


@attrs.frozen
class SequenceSelected(WidgetState):
    path: PureSequencePath
    start_time: Optional[datetime.datetime]
    shots: frozenset[PureShot]


async def get_state_async(
    sequence_path: Optional[PureSequencePath], session: AsyncExperimentSession
) -> WidgetState:
    if sequence_path is None:
        return NoSequenceSelected()
    shots_result = await session.sequences.get_shots(sequence_path)
    try:
        shots = unwrap(shots_result)
    except (PathNotFoundError, PathIsNotSequenceError):
        return NoSequenceSelected()
    else:
        start_time = unwrap(await session.sequences.get_stats(sequence_path)).start_time
    return SequenceSelected(
        path=sequence_path, shots=frozenset(shots), start_time=start_time
    )



from __future__ import annotations

from typing import (
    SupportsInt,
    Callable,
    SupportsFloat,
    overload,
)

import numpy as np

from caqtus.utils._no_public_constructor import NoPublicConstructor
from ._instructions import (
    SequencerInstruction,
    _normalize_slice,
    empty_with_dtype,
    Depth,
    Pattern,
    Array1D,
    _normalize_index,
    Length,
    Repeated,
    empty_like,
)
from ._stack import stack, merge_dtypes


def create_ramp(
    start: SupportsFloat, stop: SupportsFloat, length: SupportsInt
) -> SequencerInstruction[np.float64]:
    """Create a linear ramp between two values.

    Args:
        start: The initial value of the ramp.
        stop: The final value of the ramp.
        length: The number of points in the ramp.
    """

    length = int(length)

    start = np.float64(start)
    stop = np.float64(stop)

    if not np.isfinite(start):
        raise ValueError("Start must be finite.")

    if not np.isfinite(stop):
        raise ValueError("Stop must be finite.")

    if length < 0:
        raise ValueError("Length must be non-negative.")
    elif length == 0:
        return empty_with_dtype(start.dtype)
    else:
        return Ramp._create(start, stop, length)


class Ramp[T: (np.floating, np.void)](
    SequencerInstruction[T], metaclass=NoPublicConstructor
):
    """Represents an instruction that linearly ramps between two values.

    At index `i`, this instruction takes the value
    `start + i * (stop - start) / length`.

    Use the :func:`ramp` function to create instances of this class and don't use the
    constructor directly.

    This class is generic over the data type of the ramp.
    Since a ramp only makes sense for floating point values, the type parameter is
    constrained to be a floating point type or a structured type with only floating
    point leaf fields.
    """

    __slots__ = ("_start", "_stop", "_length")

    def __init__(self, start: T, stop: T, length: int) -> None:
        # The constructor is private. Use the :func:`ramp` function to create instances
        # of this class.
        self._start: T = start
        self._stop: T = stop
        self._length = Length(length)

        assert isinstance(self._start, np.generic)
        assert isinstance(self._stop, np.generic)
        assert Ramp.is_valid_dtype(self._start.dtype)
        assert self._start.dtype == self._stop.dtype
        assert isinstance(length, int)
        assert self._length >= 1

    @classmethod
    def is_valid_dtype(cls, dtype: np.dtype) -> bool:
        """Indicates if the dtype is valid for the start and stop values of a ramp.

        Returns:
            True if the dtype is a floating point type or a structured type with only
            floating point leaf fields, False otherwise.
        """

        if np.issubdtype(dtype, np.floating):
            return True
        if np.issubdtype(dtype, np.void):
            assert dtype.fields is not None
            return all(
                cls.is_valid_dtype(sub_dtype) for sub_dtype, *_ in dtype.fields.values()
            )
        return False

    @property
    def start(self) -> T:
        """The initial value of the ramp."""

        return self._start

    @property
    def stop(self) -> T:
        """The final value of the ramp."""

        return self._stop

    @property
    def dtype(self) -> np.dtype[T]:
        return self._start.dtype

    def __len__(self) -> Length:
        return self._length

    def __repr__(self):
        return f"{type(self).__name__}({self._start}, {self._stop}, {self._length})"

    def __str__(self):
        return f"{self._start} -{self._length}-> {self._stop}"

    @overload
    def __getitem__(self, item: int) -> T: ...

    @overload
    def __getitem__(self, item: slice) -> SequencerInstruction[T]: ...

    @overload
    def __getitem__(self, item: str) -> Ramp: ...

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get_index(item)
        elif isinstance(item, slice):
            return self._get_slice(item)
        elif isinstance(item, str):
            return self._get_channel(item)

    def _get_index(self, index: int) -> T:
        index = _normalize_index(index, len(self))

        if np.issubdtype(self.dtype, np.void):
            assert isinstance(self._start, np.void)
            assert isinstance(self._stop, np.void)
            assert self.dtype.names is not None
            value = tuple(
                self._start[name]
                + index * (self._stop[name] - self._start[name]) / self._length
                for name in self.dtype.names
            )

            return np.void(  # pyright: ignore[reportCallIssue]
                value,
                dtype=self.dtype,  # pyright: ignore[reportArgumentType]
            )
        else:
            assert isinstance(self._start, np.floating)
            assert isinstance(self._stop, np.floating)
            return self._start + index * (self._stop - self._start) / self._length

    def _get_slice(self, slice_: slice) -> SequencerInstruction[T]:
        start_index, stop_index, step = _normalize_slice(slice_, len(self))
        if step != 1:
            raise NotImplementedError
        start_value = self._get_index(start_index)
        if stop_index == len(self):
            stop_value = self._stop
        else:
            stop_value = self._get_index(stop_index)
        length = stop_index - start_index
        if length == 0:
            return empty_like(self)
        else:
            return Ramp._create(start_value, stop_value, stop_index - start_index)

    def _get_channel(self, channel: str) -> SequencerInstruction:
        if not np.issubdtype(self.dtype, np.void):
            raise ValueError("Can't get field if dtype is not a structured type.")
        assert self.dtype.names is not None
        if channel not in self.dtype.names:
            raise ValueError(f"Channel {channel} not found in dtype {self.dtype}.")
        assert isinstance(self._start, np.void)
        assert isinstance(self._stop, np.void)
        start_value = self._start[channel]
        stop_value = self._stop[channel]
        return Ramp._create(start_value, stop_value, self._length)

    def as_type[S: np.generic](self, dtype: np.dtype[S]) -> SequencerInstruction[S]:
        start = self._start.astype(dtype)
        if not isinstance(start, (np.floating, np.void)):
            raise TypeError("Can only convert to floating point or structured type.")

        stop = self._stop.astype(dtype)
        if not isinstance(stop, (np.floating, np.void)):
            raise TypeError("Can only convert to floating point or structured type.")

        return Ramp._create(start, stop, self._length)

    @property
    def depth(self) -> Depth:
        return Depth(1)

    def to_pattern(self) -> Pattern[T]:
        if np.issubdtype(self.dtype, np.void):
            assert self.dtype.names is not None
            assert isinstance(self._start, np.void)
            assert isinstance(self._stop, np.void)
            values = np.empty(len(self), dtype=self.dtype)
            for name in self.dtype.names:
                values[name] = np.linspace(
                    self._start[name], self._stop[name], len(self), endpoint=False
                )
            return Pattern.create_without_copy(values)
        else:
            assert isinstance(self._start, np.floating)
            assert isinstance(self._stop, np.floating)
            values = np.linspace(self._start, self._stop, self._length, endpoint=False)
            return Pattern.create_without_copy(values)

    def __eq__(self, other):
        if not isinstance(other, Ramp):
            return NotImplemented
        return (
            np.all(self._start == other._start)
            and np.all(self._stop == other._stop)
            and self._length == other._length
        )

    def apply[
        S: np.generic
    ](self, func: Callable[[Array1D[T]], Array1D[S]]) -> SequencerInstruction[S]:
        """Map a function element-wise to the ramp.

        Warning:
            Since an arbitrary function will not necessarily preserve linear ramps, the
            ramp is explicitly computed before applying the function.
            This may lead to poor performance and large memory usage for long ramps.
        """

        return self.to_pattern().apply(func)


@stack.register(Ramp, Ramp)
def _stack_ramps(a: Ramp, b: Ramp) -> SequencerInstruction:
    assert len(a) == len(b)

    start = _merge_values(a._start, b._start)
    stop = _merge_values(a._stop, b._stop)
    return Ramp._create(start, stop, len(a))


@stack.register(Ramp, Repeated)
def _stack_ramp_repeated(a: Ramp, b: Repeated) -> SequencerInstruction:
    if len(b.instruction) == 1:
        value = b.instruction[0]
        if not Ramp.is_valid_dtype(value.dtype):
            raise TypeError(
                f"Can't stack a ramp with an instruction having dtype {value.dtype}."
            )
        start = _merge_values(a.start, value)
        stop = _merge_values(a.stop, value)
        return Ramp._create(start, stop, len(a))
    else:
        return stack(a.to_pattern(), b.to_pattern())


@stack.register(Repeated, Ramp)
def _stack_repeated_ramp(a: Repeated, b: Ramp) -> SequencerInstruction:
    assert len(a) == len(b)

    if len(a.instruction) == 1:
        value = a.instruction[0]
        if not Ramp.is_valid_dtype(value.dtype):
            raise TypeError(
                f"Can't stack a ramp with an instruction having dtype {value.dtype}."
            )
        start = _merge_values(value, b.start)
        stop = _merge_values(value, b.stop)
        return Ramp._create(start, stop, len(b))
    else:
        return stack(a.to_pattern(), b.to_pattern())


def _merge_values(a: np.void, b: np.void) -> np.void:
    merged_dtype = merge_dtypes(a.dtype, b.dtype)
    assert a.dtype.names is not None
    assert b.dtype.names is not None
    return np.void(
        tuple(a[name] for name in a.dtype.names)
        + tuple(b[name] for name in b.dtype.names),
        dtype=merged_dtype,
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-
import os

from PySide6 import QtWidgets
from enum import Enum

from .pkg_info import __version__ as _v

__doc__ = """
| The :py:mod:`NodeGraphQt.constants` namespace contains variables and enums
 used throughout the NodeGraphQt library.
"""

# ================================== PRIVATE ===================================

MIME_TYPE = "nodegraphqt/nodes"
URI_SCHEME = "nodegraphqt://"
URN_SCHEME = "nodegraphqt::"

# PATHS
BASE_PATH = os.path.dirname(os.path.abspath(__file__))
ICON_PATH = os.path.join(BASE_PATH, "widgets", "icons")
ICON_DOWN_ARROW = os.path.join(ICON_PATH, "down_arrow.png")
ICON_NODE_BASE = os.path.join(ICON_PATH, "node_base.png")

# DRAW STACK ORDER
Z_VAL_BACKDROP = -2
Z_VAL_PIPE = -1
Z_VAL_NODE = 1
Z_VAL_PORT = 2
Z_VAL_NODE_WIDGET = 3

# ITEM CACHE MODE
# QGraphicsItem.NoCache
# QGraphicsItem.DeviceCoordinateCache
# QGraphicsItem.ItemCoordinateCache
ITEM_CACHE_MODE = QtWidgets.QGraphicsItem.DeviceCoordinateCache

# =================================== GLOBAL ===================================


class VersionEnum(Enum):
    """
    Current framework version.
    :py:mod:`NodeGraphQt.constants.VersionEnum`
    """

    #: current version string.
    VERSION = _v
    #: version major int.
    MAJOR = int(_v.split(".")[0])
    #: version minor int.
    MINOR = int(_v.split(".")[1])
    #: version patch int.
    PATCH = int(_v.split(".")[2])


class LayoutDirectionEnum(Enum):
    """
    Node graph nodes layout direction:
    :py:mod:`NodeGraphQt.constants.ViewerLayoutEnum`
    """

    #: layout nodes left to right.
    HORIZONTAL = 0
    #: layout nodes top to bottom.
    VERTICAL = 1


# =================================== VIEWER ===================================


class ViewerEnum(Enum):
    """
    Node graph viewer styling layout:
    :py:mod:`NodeGraphQt.constants.ViewerEnum`
    """

    #: default background color for the node graph.
    BACKGROUND_COLOR = (35, 35, 35)
    #: style node graph background with no grid or dots.
    GRID_DISPLAY_NONE = 0
    #: style node graph background with dots.
    GRID_DISPLAY_DOTS = 1
    #: style node graph background with grid lines.
    GRID_DISPLAY_LINES = 2
    #: grid size when styled with grid lines.
    GRID_SIZE = 50
    #: grid line color.
    GRID_COLOR = (45, 45, 45)


class ViewerNavEnum(Enum):
    """
    Node graph viewer navigation styling layout:
    :py:mod:`NodeGraphQt.constants.ViewerNavEnum`
    """

    #: default background color.
    BACKGROUND_COLOR = (25, 25, 25)
    #: default item color.
    ITEM_COLOR = (35, 35, 35)


# ==================================== NODE ====================================


class NodeEnum(Enum):
    """
    Node styling layout:
    :py:mod:`NodeGraphQt.constants.NodeEnum`
    """

    #: default node width.
    WIDTH = 160
    #: default node height.
    HEIGHT = 60
    #: default node icon size (WxH).
    ICON_SIZE = 18
    #: default node overlay color when selected.
    SELECTED_COLOR = (255, 255, 255, 30)
    #: default node border color when selected.
    SELECTED_BORDER_COLOR = (254, 207, 42, 255)


# ==================================== PORT ====================================


class PortEnum(Enum):
    """
    Port styling layout:
    :py:mod:`NodeGraphQt.constants.PortEnum`
    """

    #: default port size.
    SIZE = 22.0
    #: default port color. (r, g, b, a)
    COLOR = (49, 115, 100, 255)
    #: default port border color.
    BORDER_COLOR = (29, 202, 151, 255)
    #: port color when selected.
    ACTIVE_COLOR = (14, 45, 59, 255)
    #: port border color when selected.
    ACTIVE_BORDER_COLOR = (107, 166, 193, 255)
    #: port color on mouse over.
    HOVER_COLOR = (17, 43, 82, 255)
    #: port border color on mouse over.
    HOVER_BORDER_COLOR = (136, 255, 35, 255)
    #: threshold for selecting a port.
    CLICK_FALLOFF = 15.0


class PortTypeEnum(Enum):
    """
    Port connection types:
    :py:mod:`NodeGraphQt.constants.PortTypeEnum`
    """

    #: Connection type for input ports.
    IN = "in"
    #: Connection type for output ports.
    OUT = "out"


# ==================================== PIPE ====================================


class PipeEnum(Enum):
    """
    Pipe styling layout:
    :py:mod:`NodeGraphQt.constants.PipeEnum`
    """

    #: default width.
    WIDTH = 1.2
    #: default color.
    COLOR = (175, 95, 30, 255)
    #: pipe color to a node when it's disabled.
    DISABLED_COLOR = (200, 60, 60, 255)
    #: pipe color when selected or mouse over.
    ACTIVE_COLOR = (70, 255, 220, 255)
    #: pipe color to a node when it's selected.
    HIGHLIGHT_COLOR = (232, 184, 13, 255)
    #: draw connection as a line.
    DRAW_TYPE_DEFAULT = 0
    #: draw connection as dashed lines.
    DRAW_TYPE_DASHED = 1
    #: draw connection as a dotted line.
    DRAW_TYPE_DOTTED = 2


class PipeSlicerEnum(Enum):
    """
    Slicer Pipe styling layout:
    :py:mod:`NodeGraphQt.constants.PipeSlicerEnum`
    """

    #: default width.
    WIDTH = 1.5
    #: default color.
    COLOR = (255, 50, 75)


class PipeLayoutEnum(Enum):
    """
    Pipe connection drawing layout:
    :py:mod:`NodeGraphQt.constants.PipeLayoutEnum`
    """

    #: draw straight lines for pipe connections.
    STRAIGHT = 0
    #: draw curved lines for pipe connections.
    CURVED = 1
    #: draw angled lines for pipe connections.
    ANGLE = 2


# === PROPERTY BIN WIDGET ===


class NodePropWidgetEnum(Enum):
    """
    Mapping used for the :class:`NodeGraphQt.PropertiesBinWidget` to display a
    node property in the specified widget type.

    :py:mod:`NodeGraphQt.constants.NodePropWidgetEnum`
    """

    #: Node property will be hidden in the ``PropertiesBinWidget`` (default).
    HIDDEN = 0
    #: Node property represented with a ``QLabel`` widget.
    QLABEL = 2
    #: Node property represented with a ``QLineEdit`` widget.
    QLINE_EDIT = 3
    #: Node property represented with a ``QTextEdit`` widget.
    QTEXT_EDIT = 4
    #: Node property represented with a ``QComboBox`` widget.
    QCOMBO_BOX = 5
    #: Node property represented with a ``QCheckBox`` widget.
    QCHECK_BOX = 6
    #: Node property represented with a ``QSpinBox`` widget.
    QSPIN_BOX = 7
    #: Node property represented with a ``QDoubleSpinBox`` widget.
    QDOUBLESPIN_BOX = 8
    #: Node property represented with a ColorPicker widget.
    COLOR_PICKER = 9
    #: Node property represented with a ColorPicker (RGBA) widget.
    COLOR4_PICKER = 10
    #: Node property represented with an (Int) Slider widget.
    SLIDER = 11
    #: Node property represented with a (Dobule) Slider widget.
    DOUBLE_SLIDER = 12
    #: Node property represented with a file selector widget.
    FILE_OPEN = 13
    #: Node property represented with a file save widget.
    FILE_SAVE = 14
    #: Node property represented with a vector2 widget.
    VECTOR2 = 15
    #: Node property represented with vector3 widget.
    VECTOR3 = 16
    #: Node property represented with vector4 widget.
    VECTOR4 = 17
    #: Node property represented with float line edit widget.
    FLOAT = 18
    #: Node property represented with int line edit widget.
    INT = 19
    #: Node property represented with button widget.
    BUTTON = 20



"""Contains function for handling uncertainties in dataframes."""

from .stats import (
    compute_stats_average,
    is_error_dtype,
    get_nominal_value,
    get_error,
    VALUE_FIELD,
    ERROR_FIELD,
)

__all__ = [
    "compute_stats_average",
    "is_error_dtype",
    "get_nominal_value",
    "get_error",
    "VALUE_FIELD",
    "ERROR_FIELD",
]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'create_parameters_viewer.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class Ui_ParametersDialog(object):
    def setupUi(self, ParametersDialog):
        if not ParametersDialog.objectName():
            ParametersDialog.setObjectName("ParametersDialog")
        ParametersDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(ParametersDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.titleLabel = QLabel(ParametersDialog)
        self.titleLabel.setObjectName("titleLabel")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.titleLabel)

        self._window_title = QLineEdit(ParametersDialog)
        self._window_title.setObjectName("_window_title")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self._window_title)

        self.verticalLayout.addLayout(self.formLayout)

        self.buttonBox = QDialogButtonBox(ParametersDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(ParametersDialog)
        self.buttonBox.accepted.connect(ParametersDialog.accept)
        self.buttonBox.rejected.connect(ParametersDialog.reject)

        QMetaObject.connectSlotsByName(ParametersDialog)

    # setupUi

    def retranslateUi(self, ParametersDialog):
        ParametersDialog.setWindowTitle(
            QCoreApplication.translate("ParametersDialog", "Dialog", None)
        )
        self.titleLabel.setText(
            QCoreApplication.translate("ParametersDialog", "Title", None)
        )

    # retranslateUi



from __future__ import annotations

import datetime
from typing import Optional

import sqlalchemy
from sqlalchemy import ForeignKey, DateTime, UniqueConstraint, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ._path_table import SQLSequencePath
from ._shot_tables import SQLShot
from ._table_base import Base
from .._state import State


class SQLSequence(Base):
    __tablename__ = "sequences"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    path_id: Mapped[str] = mapped_column(
        ForeignKey(SQLSequencePath.id_, ondelete="CASCADE"), unique=True, index=True
    )
    path: Mapped[SQLSequencePath] = relationship(back_populates="sequence")
    state: Mapped[State]

    parameters: Mapped[SQLSequenceParameters] = relationship(
        cascade="all, delete",
        passive_deletes=True,
        back_populates="sequence",
    )
    iteration: Mapped[SQLIterationConfiguration] = relationship(
        cascade="all",
        passive_deletes=True,
        back_populates="sequence",
    )
    time_lanes: Mapped[SQLTimelanes] = relationship(
        cascade="all, delete",
        passive_deletes=True,
        back_populates="sequence",
    )

    device_configurations: Mapped[list[SQLDeviceConfiguration]] = relationship(
        cascade="all, delete", passive_deletes=True, back_populates="sequence"
    )

    # Stored as timezone naive datetimes, with the assumption that the timezone is UTC.
    start_time: Mapped[Optional[datetime.datetime]] = mapped_column(
        DateTime(timezone=False)
    )
    stop_time: Mapped[Optional[datetime.datetime]] = mapped_column(
        DateTime(timezone=False)
    )

    shots: Mapped[list["SQLShot"]] = relationship(
        back_populates="sequence",
        cascade="all, delete",
        passive_deletes=True,
    )
    # expected_number_of_shots indicates how many shots are expected to be executed in
    # total for this sequence.
    # It is written when the sequence iteration is set.
    # None indicates that this value is not known.
    expected_number_of_shots: Mapped[Optional[int]] = mapped_column()


class SQLSequenceParameters(Base):
    __tablename__ = "sequence.parameters"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE"), index=True
    )
    sequence: Mapped[SQLSequence] = relationship(
        back_populates="parameters", single_parent=True
    )
    content = mapped_column(sqlalchemy.types.JSON)

    __table_args__ = (UniqueConstraint(sequence_id),)


class SQLIterationConfiguration(Base):
    __tablename__ = "sequence.iteration"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE"), index=True
    )
    sequence: Mapped[SQLSequence] = relationship(
        back_populates="iteration", single_parent=True
    )
    content = mapped_column(sqlalchemy.types.JSON)

    __table_args__ = (UniqueConstraint(sequence_id),)


class SQLTimelanes(Base):
    __tablename__ = "sequence.time_lanes"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE"), index=True
    )
    sequence: Mapped[SQLSequence] = relationship(
        back_populates="time_lanes", single_parent=True
    )
    content = mapped_column(sqlalchemy.types.JSON)

    __table_args__ = (UniqueConstraint(sequence_id),)


class SQLDeviceConfiguration(Base):
    __tablename__ = "sequence.device_configurations"

    id_: Mapped[int] = mapped_column(primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE")
    )
    sequence: Mapped[SQLSequence] = relationship(back_populates="device_configurations")
    name: Mapped[str] = mapped_column(String(255))
    device_type: Mapped[str] = mapped_column(String(255))
    content = mapped_column(sqlalchemy.types.JSON)

    __table_args__ = (
        sqlalchemy.UniqueConstraint(sequence_id, name, name="device_configuration"),
    )



import attrs

from caqtus.types.image import ImageLabel
from caqtus.utils import serialization
from .timelane import TimeLane


@attrs.define
class TakePicture:
    picture_name: ImageLabel


@attrs.define(init=False, eq=False, repr=False)
class CameraTimeLane(TimeLane[TakePicture | None]):
    pass


def unstructure_hook(lane: CameraTimeLane):
    return {
        "spanned_values": serialization.unstructure(
            lane._spanned_values, list[tuple[TakePicture | None, int]]
        )
    }


def structure_hook(data, _) -> CameraTimeLane:
    structured = serialization.structure(
        data["spanned_values"], list[tuple[TakePicture | None, int]]
    )
    return CameraTimeLane.from_spanned_values(structured)


serialization.register_structure_hook(CameraTimeLane, structure_hook)
serialization.register_unstructure_hook(CameraTimeLane, unstructure_hook)



import abc
from collections.abc import MutableMapping

from caqtus.device import DeviceName, DeviceConfiguration


class DeviceConfigurationCollection(
    MutableMapping[DeviceName, DeviceConfiguration], abc.ABC
):
    """A collection of device configurations inside a session.

    This object behaves like a dictionary where the keys are the names of the devices
    and the values are the configurations of the devices.
    """

    pass



from __future__ import annotations

import abc
from typing import TypeGuard

from ..variable_name import DottedVariableName


class IterationConfiguration(abc.ABC):
    """Defines how parameters should be iterated over for a sequence.

    This is an abstract base class that defines the interface for iterations.
    It is meant to be subclassed to define different types of iterations.
    """

    @abc.abstractmethod
    def expected_number_shots(self) -> int | Unknown:
        """Return the expected number of shots defined by this iteration.

        If the number of shots can be determined ahead of time, this method should
        return that number.
        If the number of shots cannot be determined ahead of time, this method should
        return unknown.
        In doubt, the method must return unknown and not a possibly wrong guess.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_parameter_names(self) -> set[DottedVariableName]:
        """Return the names of the parameters that are iterated over.

        This method must return the name of the parameters whose values are changed
        during the iteration.
        The iteration must set the values of all these parameters before each shot.
        """

        raise NotImplementedError


class Unknown:
    """Instances of this class represent an unknown positive integer.

    It can be added or multiplied with a positive integer.
    It is absorbent for addition and for multiplication with a non-zero operand.
    Multiplication by zero return zero.
    """

    def __add__(self, other):
        match other:
            case int(x) if x >= 0:
                return self
            case Unknown():
                return self
            case _:
                return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        match other:
            case 0:
                return 0
            case Unknown():
                return self
            case int(x) if x > 0:
                return self
            case _:
                return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __repr__(self):
        return f"Unknown()"

    def __str__(self):
        return "unknown"


def is_unknown(value) -> TypeGuard[Unknown]:
    return isinstance(value, Unknown)



#!/usr/bin/python
from PySide6 import QtGui, QtCore, QtWidgets

from ..constants import Z_VAL_BACKDROP, NodeEnum
from ..qgraphics.node_abstract import AbstractNodeItem
from ..qgraphics.pipe import PipeItem
from ..qgraphics.port import PortItem


class BackdropSizer(QtWidgets.QGraphicsItem):
    """
    Sizer item for resizing a backdrop item.

    Args:
        parent (BackdropNodeItem): the parent node item.
        size (float): sizer size.
    """

    def __init__(self, parent=None, size=6.0):
        super(BackdropSizer, self).__init__(parent)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsScenePositionChanges, True)
        self.setCursor(QtGui.QCursor(QtCore.Qt.SizeFDiagCursor))
        self.setToolTip("double-click auto resize")
        self._size = size

    @property
    def size(self):
        return self._size

    def set_pos(self, x, y):
        x -= self._size
        y -= self._size
        self.setPos(x, y)

    def boundingRect(self):
        return QtCore.QRectF(0.5, 0.5, self._size, self._size)

    def itemChange(self, change, value):
        if change == QtWidgets.QGraphicsItem.ItemPositionChange:
            item = self.parentItem()
            mx, my = item.minimum_size
            x = mx if value.x() < mx else value.x()
            y = my if value.y() < my else value.y()
            value = QtCore.QPointF(x, y)
            item.on_sizer_pos_changed(value)
            return value
        return super(BackdropSizer, self).itemChange(change, value)

    def mouseDoubleClickEvent(self, event):
        item = self.parentItem()
        item.on_sizer_double_clicked()
        super(BackdropSizer, self).mouseDoubleClickEvent(event)

    def mousePressEvent(self, event):
        self.__prev_xy = (self.pos().x(), self.pos().y())
        super(BackdropSizer, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        current_xy = (self.pos().x(), self.pos().y())
        if current_xy != self.__prev_xy:
            item = self.parentItem()
            item.on_sizer_pos_mouse_release()
        del self.__prev_xy
        super(BackdropSizer, self).mouseReleaseEvent(event)

    def paint(self, painter, option, widget):
        """
        Draws the backdrop sizer in the bottom right corner.

        Args:
            painter (QtGui.QPainter): painter used for drawing the item.
            option (QtGui.QStyleOptionGraphicsItem):
                used to describe the parameters needed to draw.
            widget (QtWidgets.QWidget): not used.
        """
        painter.save()

        margin = 1.0
        rect = self.boundingRect()
        rect = QtCore.QRectF(
            rect.left() + margin,
            rect.top() + margin,
            rect.width() - (margin * 2),
            rect.height() - (margin * 2),
        )

        item = self.parentItem()
        if item and item.selected:
            color = QtGui.QColor(*NodeEnum.SELECTED_BORDER_COLOR.value)
        else:
            color = QtGui.QColor(*item.color)
            color = color.darker(110)
        path = QtGui.QPainterPath()
        path.moveTo(rect.topRight())
        path.lineTo(rect.bottomRight())
        path.lineTo(rect.bottomLeft())
        painter.setBrush(color)
        painter.setPen(QtCore.Qt.NoPen)
        painter.fillPath(path, painter.brush())

        painter.restore()


class BackdropNodeItem(AbstractNodeItem):
    """
    Base Backdrop item.

    Args:
        name (str): name displayed on the node.
        text (str): backdrop text.
        parent (QtWidgets.QGraphicsItem): parent item.
    """

    def __init__(self, name="backdrop", text="", parent=None):
        super(BackdropNodeItem, self).__init__(name, parent)
        self.setZValue(Z_VAL_BACKDROP)
        self._properties["backdrop_text"] = text
        self._min_size = 80, 80
        self._sizer = BackdropSizer(self, 26.0)
        self._sizer.set_pos(*self._min_size)
        self._nodes = [self]

    def _combined_rect(self, nodes):
        group = self.scene().createItemGroup(nodes)
        rect = group.boundingRect()
        self.scene().destroyItemGroup(group)
        return rect

    def mouseDoubleClickEvent(self, event):
        viewer = self.viewer()
        if viewer:
            viewer.node_double_clicked.emit(self.id)
        super(BackdropNodeItem, self).mouseDoubleClickEvent(event)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            pos = event.scenePos()
            rect = QtCore.QRectF(pos.x() - 5, pos.y() - 5, 10, 10)
            item = self.scene().items(rect)[0]

            if isinstance(item, (PortItem, PipeItem)):
                self.setFlag(self.GraphicsItemFlag.ItemIsMovable, False)
                return
            if self.selected:
                return

            viewer = self.viewer()
            [n.setSelected(False) for n in viewer.selected_nodes()]

            self._nodes += self.get_nodes(False)
            [n.setSelected(True) for n in self._nodes]

    def mouseReleaseEvent(self, event):
        super(BackdropNodeItem, self).mouseReleaseEvent(event)
        self.setFlag(self.GraphicsItemFlag.ItemIsMovable, True)
        [n.setSelected(True) for n in self._nodes]
        self._nodes = [self]

    def on_sizer_pos_changed(self, pos):
        self._width = pos.x() + self._sizer.size
        self._height = pos.y() + self._sizer.size

    def on_sizer_pos_mouse_release(self):
        size = {"pos": self.xy_pos, "width": self._width, "height": self._height}
        self.viewer().node_backdrop_updated.emit(self.id, "sizer_mouse_release", size)

    def on_sizer_double_clicked(self):
        size = self.calc_backdrop_size()
        self.viewer().node_backdrop_updated.emit(self.id, "sizer_double_clicked", size)

    def paint(self, painter, option, widget):
        """
        Draws the backdrop rect.

        Args:
            painter (QtGui.QPainter): painter used for drawing the item.
            option (QtGui.QStyleOptionGraphicsItem):
                used to describe the parameters needed to draw.
            widget (QtWidgets.QWidget): not used.
        """
        painter.save()
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtCore.Qt.NoBrush)

        margin = 1.0
        rect = self.boundingRect()
        rect = QtCore.QRectF(
            rect.left() + margin,
            rect.top() + margin,
            rect.width() - (margin * 2),
            rect.height() - (margin * 2),
        )

        radius = 2.6
        color = (self.color[0], self.color[1], self.color[2], 50)
        painter.setBrush(QtGui.QColor(*color))
        painter.setPen(QtCore.Qt.NoPen)
        painter.drawRoundedRect(rect, radius, radius)

        top_rect = QtCore.QRectF(rect.x(), rect.y(), rect.width(), 26.0)
        painter.setBrush(QtGui.QBrush(QtGui.QColor(*self.color)))
        painter.setPen(QtCore.Qt.NoPen)
        painter.drawRoundedRect(top_rect, radius, radius)
        for pos in [top_rect.left(), top_rect.right() - 5.0]:
            painter.drawRect(QtCore.QRectF(pos, top_rect.bottom() - 5.0, 5.0, 5.0))

        if self.backdrop_text:
            painter.setPen(QtGui.QColor(*self.text_color))
            txt_rect = QtCore.QRectF(
                top_rect.x() + 5.0,
                top_rect.height() + 3.0,
                rect.width() - 5.0,
                rect.height(),
            )
            painter.setPen(QtGui.QColor(*self.text_color))
            painter.drawText(
                txt_rect,
                QtCore.Qt.AlignLeft | QtCore.Qt.TextWordWrap,
                self.backdrop_text,
            )

        if self.selected:
            sel_color = [x for x in NodeEnum.SELECTED_COLOR.value]
            sel_color[-1] = 15
            painter.setBrush(QtGui.QColor(*sel_color))
            painter.setPen(QtCore.Qt.NoPen)
            painter.drawRoundedRect(rect, radius, radius)

        txt_rect = QtCore.QRectF(
            top_rect.x(), top_rect.y(), rect.width(), top_rect.height()
        )
        painter.setPen(QtGui.QColor(*self.text_color))
        painter.drawText(txt_rect, QtCore.Qt.AlignCenter, self.name)

        border = 0.8
        border_color = self.color
        if self.selected and NodeEnum.SELECTED_BORDER_COLOR.value:
            border = 1.0
            border_color = NodeEnum.SELECTED_BORDER_COLOR.value
        painter.setBrush(QtCore.Qt.NoBrush)
        painter.setPen(QtGui.QPen(QtGui.QColor(*border_color), border))
        painter.drawRoundedRect(rect, radius, radius)

        painter.restore()

    def get_nodes(self, inc_intersects=False):
        mode = {True: QtCore.Qt.IntersectsItemShape, False: QtCore.Qt.ContainsItemShape}
        nodes = []
        if self.scene():
            polygon = self.mapToScene(self.boundingRect())
            rect = polygon.boundingRect()
            items = self.scene().items(rect, mode=mode[inc_intersects])
            for item in items:
                if item == self or item == self._sizer:
                    continue
                if isinstance(item, AbstractNodeItem):
                    nodes.append(item)
        return nodes

    def calc_backdrop_size(self, nodes=None):
        nodes = nodes or self.get_nodes(True)
        if nodes:
            nodes_rect = self._combined_rect(nodes)
        else:
            center = self.mapToScene(self.boundingRect().center())
            nodes_rect = QtCore.QRectF(
                center.x(), center.y(), self._min_size[0], self._min_size[1]
            )

        padding = 40
        return {
            "pos": [nodes_rect.x() - padding, nodes_rect.y() - padding],
            "width": nodes_rect.width() + (padding * 2),
            "height": nodes_rect.height() + (padding * 2),
        }

    @property
    def minimum_size(self):
        return self._min_size

    @minimum_size.setter
    def minimum_size(self, size=(50, 50)):
        self._min_size = size

    @property
    def backdrop_text(self):
        return self._properties["backdrop_text"]

    @backdrop_text.setter
    def backdrop_text(self, text):
        self._properties["backdrop_text"] = text
        self.update(self.boundingRect())

    @AbstractNodeItem.width.setter
    def width(self, width=0.0):
        AbstractNodeItem.width.fset(self, width)
        self._sizer.set_pos(self._width, self._height)

    @AbstractNodeItem.height.setter
    def height(self, height=0.0):
        AbstractNodeItem.height.fset(self, height)
        self._sizer.set_pos(self._width, self._height)

    def from_dict(self, node_dict):
        super().from_dict(node_dict)
        custom_props = node_dict.get("custom") or {}
        for prop_name, value in custom_props.items():
            if prop_name == "backdrop_text":
                self.backdrop_text = value



from __future__ import annotations

import abc
from typing import Optional

import polars
from PyQt6.QtWidgets import QWidget

import caqtus.gui.qtutil.qabc as qabc


class ViewCreator(qabc.QABC):
    """Interface for a class the creates a new view over data.

    In the graphplot application, the user can select which view creator to use for creating a new view. If a view
    creator is selected by the user, the graphplot main window will call the create_view method of the view creator and
    set the returned view its central widget.
    If the view creator is a QWidget, it will be displayed in the main window in the view creator settings group so that
    users can interact with the view creator. If the view creator is not a QWidget, the settings group will be hidden.
    """

    @abc.abstractmethod
    def create_view(self) -> DataView:
        """Create a new view.

        This method will be called when the user selects a new view in the graphplot main window. The view should be
        created and returned. The view will be displayed in the main window. The main window will take ownership of the
        view.
        """

        raise NotImplementedError()


class DataView(QWidget, qabc.QABC):
    """Interface for widget that displays of view over data.

    The main widget of the graphplot application is a DataView. User can select which DataView to display in the main
    window. The DataView is responsible for displaying the data in a meaningful way.
    """

    @abc.abstractmethod
    def update_data(self, dataframe: Optional[polars.DataFrame]) -> None:
        """Update the view with the given data.

        This method will be called automatically by the graphplot main window when new data is available. The data is
        provided as a polars DataFrame, which is a table-like data structure.
        This method will always be called from the main GUI thread, so it is safe to update the GUI from this method.
        However, it is recommended to make this method as fast as possible, otherwise the GUI will become unresponsive.
        """

        raise NotImplementedError()



import abc
from typing import Protocol

from caqtus.utils.serialization import JSON
from ..timelane import TimeLane, TimeLanes


class TimeLaneSerializerProtocol(Protocol):
    """Defines how to (un)structure time lanes."""

    @abc.abstractmethod
    def dump(self, lane: TimeLane) -> JSON: ...

    @abc.abstractmethod
    def load(self, data: JSON) -> TimeLane: ...

    @abc.abstractmethod
    def unstructure_time_lanes(self, time_lanes: TimeLanes) -> JSON: ...

    @abc.abstractmethod
    def structure_time_lanes(self, content: JSON) -> TimeLanes: ...



import datetime
from typing import TYPE_CHECKING

from sqlalchemy import UniqueConstraint, ForeignKey, JSON, LargeBinary, DateTime, String
from sqlalchemy.orm import mapped_column, Mapped, relationship

from ._table_base import Base

if TYPE_CHECKING:
    from ._sequence_table import SQLSequence


class SQLShot(Base):
    __tablename__ = "shots"

    __table_args__ = (UniqueConstraint("sequence_id", "index", name="shot_identifier"),)

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey("sequences.id", ondelete="CASCADE"), index=True
    )
    sequence: Mapped["SQLSequence"] = relationship(back_populates="shots")
    index: Mapped[int] = mapped_column(index=True)

    # Stored without explicit timezone, with the assumption that all times are UTC.
    start_time: Mapped[datetime.datetime] = mapped_column(DateTime(timezone=False))
    end_time: Mapped[datetime.datetime] = mapped_column(DateTime(timezone=False))

    parameters: Mapped["SQLShotParameter"] = relationship(
        back_populates="shot",
        cascade="all, delete",
        passive_deletes=True,
    )
    array_data: Mapped[list["SQLShotArray"]] = relationship(
        back_populates="shot",
        cascade="all, delete",
        passive_deletes=True,
    )
    structured_data: Mapped[list["SQLStructuredShotData"]] = relationship(
        back_populates="shot",
        cascade="all, delete",
        passive_deletes=True,
    )


# Shot parameters are stored as a single JSON row per shot, which is a dictionary
# mapping parameter names to values.
# Since we usually load all parameters at once, there is not much need to have a row for
# each single parameter.
class SQLShotParameter(Base):
    __tablename__ = "shot.parameters"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    shot_id: Mapped[int] = mapped_column(
        ForeignKey("shots.id", ondelete="CASCADE"),
        unique=True,
        index=True,
    )
    shot: Mapped[SQLShot] = relationship(back_populates="parameters")
    content = mapped_column(JSON)


# Unlike parameters, data are stored as many rows per shot. This is because we might not
# need to load all data for each shot every time.
# Data are separated in two sub-tables: one for structured nested data that contains
# few elements, and one for large arrays of bytes.
class SQLStructuredShotData(Base):
    __tablename__ = "shot.data.structured"

    __table_args__ = (
        UniqueConstraint("shot_id", "label", name="structured_identifier"),
    )

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    shot_id: Mapped[int] = mapped_column(
        ForeignKey("shots.id", ondelete="CASCADE"),
        index=True,
    )
    shot: Mapped[SQLShot] = relationship(back_populates="structured_data")
    label: Mapped[str] = mapped_column(String(255))
    content = mapped_column(JSON)


class SQLShotArray(Base):
    __tablename__ = "shot.data.arrays"

    __table_args__ = (UniqueConstraint("shot_id", "label", name="array_identifier"),)

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    shot_id: Mapped[int] = mapped_column(
        ForeignKey("shots.id", ondelete="CASCADE"),
        index=True,
    )
    shot: Mapped[SQLShot] = relationship(back_populates="array_data")
    label: Mapped[str] = mapped_column(String(255))
    dtype: Mapped[str] = mapped_column(String(255))
    shape = mapped_column(JSON)
    bytes_ = mapped_column(LargeBinary, name="bytes")  # stored in C-order



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'create_atom_viewer.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class Ui_AtomDialog(object):
    def setupUi(self, AtomDialog):
        if not AtomDialog.objectName():
            AtomDialog.setObjectName("AtomDialog")
        AtomDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(AtomDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.titleLabel = QLabel(AtomDialog)
        self.titleLabel.setObjectName("titleLabel")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.titleLabel)

        self._window_title = QLineEdit(AtomDialog)
        self._window_title.setObjectName("_window_title")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self._window_title)

        self.cameraLabel = QLabel(AtomDialog)
        self.cameraLabel.setObjectName("cameraLabel")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.cameraLabel)

        self._detector_line_edit = QLineEdit(AtomDialog)
        self._detector_line_edit.setObjectName("_detector_line_edit")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self._detector_line_edit)

        self.pictureLabel = QLabel(AtomDialog)
        self.pictureLabel.setObjectName("pictureLabel")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.pictureLabel)

        self._picture_line_edit = QLineEdit(AtomDialog)
        self._picture_line_edit.setObjectName("_picture_line_edit")

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self._picture_line_edit)

        self.axesEqualLabel = QLabel(AtomDialog)
        self.axesEqualLabel.setObjectName("axesEqualLabel")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.axesEqualLabel)

        self.axesEqualCheckBox = QCheckBox(AtomDialog)
        self.axesEqualCheckBox.setObjectName("axesEqualCheckBox")
        self.axesEqualCheckBox.setChecked(True)

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.axesEqualCheckBox)

        self.verticalLayout.addLayout(self.formLayout)

        self.buttonBox = QDialogButtonBox(AtomDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(AtomDialog)
        self.buttonBox.accepted.connect(AtomDialog.accept)
        self.buttonBox.rejected.connect(AtomDialog.reject)

        QMetaObject.connectSlotsByName(AtomDialog)

    # setupUi

    def retranslateUi(self, AtomDialog):
        AtomDialog.setWindowTitle(
            QCoreApplication.translate("AtomDialog", "Dialog", None)
        )
        self.titleLabel.setText(QCoreApplication.translate("AtomDialog", "Title", None))
        self.cameraLabel.setText(
            QCoreApplication.translate("AtomDialog", "Detector", None)
        )
        self.pictureLabel.setText(
            QCoreApplication.translate("AtomDialog", "Picture", None)
        )
        self.axesEqualLabel.setText(
            QCoreApplication.translate("AtomDialog", "Axes equal", None)
        )

    # retranslateUi



from typing import Optional

from PySide6.QtCore import QModelIndex, Qt, QObject
from PySide6.QtGui import (
    QColor,
    QPalette,
    QPainter,
)
from PySide6.QtWidgets import (
    QStyleOptionViewItem,
    QStyleOptionProgressBar,
    QApplication,
    QStyle,
    QStyledItemDelegate,
)

from caqtus.session import State
from caqtus.session._sequence_collection import SequenceStats
from caqtus.types.iteration import is_unknown


class ProgressDelegate(QStyledItemDelegate):
    """Delegate to display the progress of a sequence.

    This custom delegate can be used to display the progress of a sequence in a
    view.
    It should be used for indices that have a :class:`SequenceStats` as their
    display role.
    """

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._progress_bar_option = QStyleOptionProgressBar()
        self._progress_bar_option.textVisible = True
        self._progress_bar_option.minimum = 0

        self._progress_bar_colors = {
            State.DRAFT: None,
            State.PREPARING: None,
            State.RUNNING: None,
            State.INTERRUPTED: QColor(166, 138, 13),
            State.FINISHED: QColor(98, 151, 85),
            State.CRASHED: QColor(240, 82, 79),
        }

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        sequence_stats = index.data(Qt.ItemDataRole.DisplayRole)
        if isinstance(sequence_stats, SequenceStats):
            self._progress_bar_option.rect = option.rect
            self._progress_bar_option.palette = option.palette

            # Need to choose the color of the text that overlays the progress bar.
            if option.state & QStyle.StateFlag.State_Selected:
                color = option.palette.color(QPalette.ColorRole.HighlightedText)
            else:
                color = option.palette.color(QPalette.ColorRole.Text)
            self._progress_bar_option.palette.setColor(
                QPalette.ColorRole.HighlightedText, color
            )
            progress, maximum = self._get_progress_and_max(sequence_stats)
            self._progress_bar_option.progress = progress
            self._progress_bar_option.maximum = maximum
            state = sequence_stats.state
            self._progress_bar_option.text = self._get_text(state)
            text_color = self._get_text_color(state)
            if text_color is not None:
                self._progress_bar_option.palette.setColor(
                    QPalette.ColorRole.Text, text_color
                )
            highlight_color = self._get_highlight_color(state)
            if highlight_color is not None:
                self._progress_bar_option.palette.setColor(
                    QPalette.ColorRole.Highlight, highlight_color
                )
            QApplication.style().drawControl(
                QStyle.ControlElement.CE_ProgressBar, self._progress_bar_option, painter
            )
        else:
            super().paint(painter, option, index)

    @staticmethod
    def _get_text(state: State) -> str:
        return state.value

    def _get_text_color(self, state: State) -> Optional[QColor]:
        return self._get_highlight_color(state)

    def _get_highlight_color(self, state: State) -> Optional[QColor]:
        return self._progress_bar_colors[state]

    @staticmethod
    def _get_progress_and_max(sequence_stats: SequenceStats) -> tuple[int, int]:
        state = sequence_stats.state
        if state == State.DRAFT or state == State.PREPARING:
            progress = 0
            maximum = 100
        else:
            total = sequence_stats.expected_number_shots
            if not is_unknown(total):
                progress = sequence_stats.number_completed_shots
                maximum = total
            else:
                if state == State.RUNNING:  # in progress bar
                    progress = 0
                    maximum = 0
                else:  # filled fixed bar
                    progress = 1
                    maximum = 1
        return progress, maximum



"""This module provides widgets to edit a sequencer configuration."""

from ._trigger_selector import TriggerSelector
from .channels_widget import SequencerChannelWidget
from .sequencer_configuration_editor import SequencerConfigurationEditor, TimeStepEditor
from .channel_output_editor import ChannelOutputEditor

__all__ = [
    "SequencerConfigurationEditor",
    "TimeStepEditor",
    "SequencerChannelWidget",
    "TriggerSelector",
    "ChannelOutputEditor",
]



"""Contains classes and functions to manage devices."""

from ._controller import DeviceController
from .configuration import (
    DeviceConfiguration,
    DeviceParameter,
)
from ._name import DeviceName
from .runtime import Device, RuntimeDevice
from . import sequencer
from . import camera
from . import output_transform

__all__ = [
    "DeviceName",
    "DeviceConfiguration",
    "DeviceParameter",
    "Device",
    "RuntimeDevice",
    "DeviceController",
    "sequencer",
    "camera",
    "output_transform",
]



from __future__ import annotations

from abc import abstractmethod
from typing import Optional, TypeVar, Generic, Callable, NewType

import attrs
from PySide6.QtWidgets import QWidget

import caqtus.gui.qtutil.qabc as qabc
from caqtus.session import Shot
from caqtus.utils.serialization import JSON


class ShotView(QWidget, metaclass=qabc.QABCMeta):
    @abstractmethod
    async def display_shot(self, shot: Shot) -> None:
        raise NotImplementedError


S = TypeVar("S", bound=JSON)
V = TypeVar("V", bound=ShotView)

ManagerName = NewType("ManagerName", str)


@attrs.define
class ViewManager(Generic[V, S]):
    constructor: Callable[[S], V]
    dumper: Callable[[V], S]
    state_generator: Callable[[QWidget], Optional[tuple[str, S]]]
